/**
  *  \file arch/generic.h
  *  \brief Generic Platform Definition
  */
#ifndef PCC_V2_ARCH_GENERIC_H
#define PCC_V2_ARCH_GENERIC_H

#include "cpluslib/string.h"

// Forward references:
class StandardCommandLineParser;
void enableStdioConsole();

/** Generic platform definition. This is a template for implementation
    of the actual platform's Platform class. It can be used as a base
    class (but doesn't have to). Functions not implemented in the
    Platform class will then be found in this one. This does not use
    virtual functions, because there will be only one Platform class
    active in any compilation, so all references can be resolved at
    compile time, and all empty platform hooks can be completely
    optimized away. */
struct GenericPlatform {
    /** Post-graphics initialisation hook. A program that uses
        graphics mode must call postGraphicsInit() after SDL_Init().
        Rationale: graphics programs must be setuid for SVGAlib.
        SVGAlib correctly drops root privileges when started.
        When using another graphics backend, we may still be setuid,
        but do not want to.

        No such handling is required for text mode programs. If users
        want these setuid, we will not stop them. */
    static inline void postGraphicsInit() { }

    /** Initialize platform-specific stuff. Call this at program
        start. Unimplemented, platform code MUST override it. */
    static void initPlatform();

    /** Initialize options. Call this before parameter parsing. */
    static inline void initOptions(StandardCommandLineParser&) { }

    /** Initialize output, for console programs. Must hook something
        to the console which causes output to get visible. */
    static void initConsoleOutput();

    /** Initialize output, for GUI programs. This may do the same as
        initConsoleOutput() on operating systems like Unix where GUI
        programs can have a console. On Windows, it will do nothing,
        but there will be a meaningful showOutput(). */
    static void initGuiOutput();

    /** Console input. Same thing as std::getline(), but must produce
        input in UTF-8. To be called from console programs only. */
    static bool readConsoleLine(string_t prompt, string_t& input);

    /** Show output. Call this before exiting when the GUI was not
        run, from a GUI program. */
    static void showOutput();

    /** Get data directory (planets root). Unimplemented, platform
        code MUST override it. */
    static string_t getDataDirectory();

    /** Get resource directory. Unimplemented, platform code MUST
        override it. */
    static string_t getResourceDirectory();

    /** Get user home directory. Unimplemented, platform code MUST
        override it. */
    static string_t getUserSettingsDirectory();

    /** Get language code. */
    static string_t getLanguageCode() { return string_t(); }

    /** Convert arguments to UTF8. Unimplemented, platform code
        MUST override it. */
    static char** convertArgumentsToUtf8(char** argv);

    /** Free argument strings converted by convertArgumentsToUtf8. */
    static void freeArguments(char** argv);

    /** Path separator character. */
    static const char PATH_SEPARATOR = '/';

    /** Line separator sequence. Unimplemented, platform code MUST
        override it. */
    static const char LINE_SEPARATOR[];
};

#endif
