/**
  *  \file arch/win32/file.cc
  *  \brief File I/O based upon Win32 file handles
  */

#include <windows.h>
#include "io/streams.h"
#include "arch/win32/win32.h"
#include "util/nls.h"
#include "util/except.h"

namespace {
    /** \class Win32FileStream
        \brief File I/O using a Win32 file handle */
    class Win32FileStream : public Stream {
        /** File descriptor. */
        HANDLE hFile;
        /** File name. UTF-8 encoding. */
        const string_t name;

        void init(const string_t utf8Name, int mode);
     public:
        Win32FileStream(string_t name, int mode);
        ~Win32FileStream();
    
        std::size_t read(char* buffer, std::size_t max);
        std::size_t write(const char* buffer, std::size_t max);
        int getCapabilities() const;
        bool seek(fpos_t pos);
        fpos_t getPos();
        fpos_t getSize();
        string_t getName();
    };
}

Ptr<Stream>
Stream::openFile(const string_t& name, int mode)
{
    return new Win32FileStream(name, mode);
}

/** Constructor.
    \param name File name in UTF-8 encoding.
    \param mode Open mode */
Win32FileStream::Win32FileStream(const string_t name, const int mode)
    : hFile(INVALID_HANDLE_VALUE),
      name(name)
{
    init(name, mode);
}

Win32FileStream::~Win32FileStream()
{
    if (hFile != INVALID_HANDLE_VALUE)
        CloseHandle(hFile);
}

/** Open file. This converts the mode and invokes the actual system call.
    It sets \c hFile accordingly. It throws on error.
    \param utf8Name UTF-8 name
    \param mode Mode (C_xxx) */
void
Win32FileStream::init(const string_t utf8Name, const int mode)
{
    // dwDesiredAccess parameter:
    int desiredAccess = GENERIC_READ;
    if ((mode & (C_CREATE | C_WRITE)) != 0)
        desiredAccess |= GENERIC_WRITE;

    // dwCreationDistribution parameter:
    int creationDistribution;
    if (mode & C_CREATE) {
        creationDistribution = CREATE_ALWAYS;
    } else {
        creationDistribution = OPEN_EXISTING;
    }

    // Open file
    if (win32::hasUnicodeSupport()) {
        win32::WStr sysName;
        win32::convertToUnicode(sysName, utf8Name, true);
        hFile = CreateFileW(&sysName[0], desiredAccess, 0, 0, creationDistribution, FILE_ATTRIBUTE_NORMAL, 0);
    } else {
        string_t sysName = win32::convertToANSI(utf8Name);
        hFile = CreateFileA(sysName.c_str(), desiredAccess, 0, 0, creationDistribution, FILE_ATTRIBUTE_NORMAL, 0);
    }

    // Regular error return is INVALID_HANDLE_VALUE. Win95b returns 0 from
    // CreateFileW, so check for both.
    if (hFile == 0 || hFile == INVALID_HANDLE_VALUE) {
        throw FileOpenException(utf8Name, win32::getLastErrorString());
    }
}

std::size_t
Win32FileStream::read(char* buffer, std::size_t max)
{
    DWORD got;
    if (ReadFile(hFile, buffer, max, &got, 0))
        return got;
    else
        return 0;
}

std::size_t
Win32FileStream::write(const char* buffer, std::size_t max)
{
    DWORD written;
    if (WriteFile(hFile, buffer, max, &written, 0))
        return written;
    else
        return 0;
}

bool
Win32FileStream::seek(fpos_t pos)
{
    return SetFilePointer(hFile, pos, 0, FILE_BEGIN) != 0xFFFFFFFF;
}

Win32FileStream::fpos_t
Win32FileStream::getPos()
{
    return SetFilePointer(hFile, 0, 0, FILE_CURRENT);
}

Win32FileStream::fpos_t
Win32FileStream::getSize()
{
    return GetFileSize(hFile, 0);
}

int
Win32FileStream::getCapabilities() const
{
    return C_READ | C_WRITE | C_SEEK;
}

string_t
Win32FileStream::getName()
{
    return name;
}
