/**
  *  \file game/coord.cc
  *  \brief Game Coordinate System Handling
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <ostream>
#include "game/coord.h"
#include "game/config.h"
#include "game/host.h"
#include "util/misc.h"
#include "util/modinit.h"

/** Global starchart configuration. */
GChartConfiguration chart_conf = {
    false,
    1000, 1000,                 // min
    3000, 3000,                 // max
    2000, 2000                  // size
};

/********************************* GPoint ********************************/

/** Check if point is on map.
    Consistent with behaviour of PCC 1.x, PHost 3.3c, and EchoView. */
bool
GPoint::isOnMap() const
{
    if (!chart_conf.wrapped) {
        return true;
    } else {
        return (x < chart_conf.xmax && x >= chart_conf.xmin
                && y < chart_conf.ymax && y >= chart_conf.ymin);
    }
}

/** Get normalized version of this point. This maps "outside" points into
    "inside" points, and thus returns the incarnation of this point that is
    on the main tile of the map. */
GPoint
GPoint::getNormalized() const
{
    int x = this->x, y = this->y;
    if (chart_conf.wrapped) {
        if (x >= chart_conf.xmax)
            x -= chart_conf.xsize;
        if (x <  chart_conf.xmin)
            x += chart_conf.xsize;
        if (y >= chart_conf.ymax)
            y -= chart_conf.ysize;
        if (y <  chart_conf.ymin)
            y += chart_conf.ysize;
    }
    return GPoint(x, y);
}

/** Get incarnation nearest to a. On a wrapped map, returns
    the incarnation of this point that is nearest to a.
    Usage: player clicks into map at point a, and we find that
    planet X is near to it. The equivalent coordinates are
    X.getPos().getNearestTo(a), which accounts for the fact
    that the player may have clicked at the "wrong" side of the map
    border. */
GPoint
GPoint::getNearestTo(GPoint a) const
{
    int x = this->x, y = this->y;
    if (chart_conf.wrapped) {
        if (2*(a.x - x) > chart_conf.xsize)
            x += chart_conf.xsize;
        if (2*(a.y - y) > chart_conf.ysize)
            y += chart_conf.ysize;
        if (2*(x - a.x) > chart_conf.xsize)
            x -= chart_conf.xsize;
        if (2*(y - a.y) > chart_conf.ysize)
            y -= chart_conf.ysize;
    }
    return GPoint(x, y);
}

/** Set this point by parsing a coordinate pair.
    \param str Coordinate
    \return true on success */
bool
GPoint::parseCoordinates(const string_t& str)
{
    int x, y;
    string_t::size_type p = str.find(',');
    if (p == str.npos)
        return false;

    if (!stringValue(str.substr(0, p), x) || !stringValue(str.substr(p+1), y))
        return false;

    this->x = x;
    this->y = y;
    return true;
}

/** Set this point by parsing a sector number.
    \param s Sector number as string
    \return true on success */
bool
GPoint::parseSectorNumber(const string_t& s)
{
    int sec;
    return stringValue(s, sec)
        && parseSectorNumber(sec);
}

/** Set this point by parsing a sector number.
    \param n Sector number as integer
    \return true on success */
bool
GPoint::parseSectorNumber(int n)
{
    /* Valid range is 100 .. 499 */
    if (n < 100 || n > 499)
        return false;

    /* Compute location */
    static const int secX[] = { -950,   50, -950,   50 };
    static const int secY[] = {  950,  950,  -50,  -50 };
    int originX = (chart_conf.xmax + chart_conf.xmin) / 2;
    int originY = (chart_conf.ymax + chart_conf.ymin) / 2;
    int newX = originX + secX[n / 100 - 1] + 100*((n % 100) / 10);
    int newY = originY + secY[n / 100 - 1] - 100*(n % 10);

    /* On map? */
    if (!GPoint(newX, newY).isOnMap())
        return false;

    /* Use it */
    this->x = newX;
    this->y = newY;

    return true;
}

/** Get sector number this point is in, as decimal number. This
    returns the same sector number as in PCC 1.x, which happens to be
    the same as the sectors on Trevor Fuson's VGAMAP for a
    standard-sized map.
    \return sector number (100..499), zero if point is not in any
    numbered sector */
int
GPoint::getSectorNumber() const
{
    /* outside map? */
    if (!isOnMap())
        return 0;

    int x = this->x - (chart_conf.xmax + chart_conf.xmin) / 2 + 1000;
    int y = this->y - (chart_conf.ymax + chart_conf.ymin) / 2 + 1000;

    /* outside known region? */
    if (x < 0 || y < 0 || x >= 2000 || y >= 2000)
        return 0;

    int major = 1;
    if (x >= 1000)
        major += 1;
    if (y < 1000)
        major += 2;

    return 100 * major
        + 10 * ((x % 1000) / 100)
        + (999 - (y % 1000)) / 100;
}

/************************** GChartConfiguration **************************/

/** Initialize configuration. Validates the configuration and computes
    the derived information. */
void
GChartConfiguration::init()
{
    if (xmax < xmin)
        std::swap(xmax, xmin);
    if (ymax < xmin)
        std::swap(ymax, ymin);

    xsize = xmax - xmin;
    ysize = ymax - ymin;
}

/** Get number of map images. */
int
GChartConfiguration::getNumMapImages() const
{
    if (wrapped)
        return 9;
    else
        return 1;
}

/** Get alias of a point in another map image. */
GPoint
GChartConfiguration::getPointAlias(GPoint pt, int nr) const
{
    if (wrapped) {
        /* Images:
             1 2 3
             4 0 5
             6 7 8 */
        static const int8 dx[] = {  0, -1,  0, +1, -1, +1, -1,  0, +1 };
        static const int8 dy[] = {  0, -1, -1, -1,  0,  0, +1, +1, +1 };
        pt.x += dx[nr] * xsize;
        pt.y += dy[nr] * ysize;
    }
    return pt;
}

/******************************* Utilities *******************************/

/** Get minimum distance between two points, squared. */
long
distanceSquared(GPoint a, GPoint b)
{
    GPoint c = b.getNearestTo(a);
    return distanceSquaredRaw(a, c);
}

/** Get distance between two points, squared. */
long
distanceSquaredRaw(GPoint a, GPoint b)
{
    long x = a.x - b.x;
    long y = a.y - b.y;
    return x*x + y*y;
}

/** Initialize. Assumes configuration has been read. */
static void
initCoordinates()
{
    if (host.isPHost()) {
        chart_conf.wrapped = config.AllowWraparoundMap();
        chart_conf.xmin = config.WraparoundRectangle(1);
        chart_conf.ymin = config.WraparoundRectangle(2);
        chart_conf.xmax = config.WraparoundRectangle(3);
        chart_conf.ymax = config.WraparoundRectangle(4);
    }
    chart_conf.init();
}

/** Output GPoint to a stream. */
std::ostream&
operator<<(std::ostream& os, GPoint pt)
{
    return os << '(' << pt.x << ", " << pt.y << ')';
}

// Constraints: after mod_config
const ModuleInitializer mod_coord[] = {
    // Initialize this module whenever configuration has been read.
    // Configuration is modified in three cases:
    { OnStartup | OnEnterNoGame | OnEnterGame, initCoordinates },
    { 0, 0 }
};
