/**
  *  \file game/mission.h
  *  \brief Starship Mission Management
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_MISSION_H
#define PCC_V2_GAME_MISSION_H

#include <vector>
#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "game/struct.h"
#include "game/playerset.h"
#include "util/modinit.h"

class GShip;
class Stream;

/** Mission description. Describes a ship mission and its arguments.
    A mission is identified by its number and race mask. */
class GMission {
    int        number;           ///< Mission.
    GPlayerSet race_mask;        ///< Race mask.
    unsigned   arg_spec[2];      ///< Argument definitions. ma_XXXX, indexed by Parameter.
    unsigned   flags;            ///< Other flags. mf_XXX.
    string_t   name;             ///< Mission name used in selection lists.
    string_t   short_name;       ///< Short name. Used for scripts.
    string_t   arg_names[2];     ///< Argument names. Indexed by Parameter.
    string_t   exp_condition;    ///< Condition for selectability.
    string_t   exp_warning;      ///< Condition to check whether it will work.
    string_t   exp_label;        ///< Expression for label on ship screen.
    string_t   cmd_onset;        ///< "on set" expression.
    char       hotkey;           ///< Mission hot-key.

    void       parseDescription(string_t desc_line);
 public:
    /** Parameter identification for a mission. */
    enum Parameter {
        InterceptParameter,      ///< Intercept parameter.
        TowParameter             ///< Tow parameter.
    };

    GMission();
    GMission(int number, string_t desc_line);
    ~GMission();

    /** Get mission number. */
    int getNumber() const
        { return number; }
    /** Get races which can do this mission.
        \return set of race numbers */
    GPlayerSet getRaceMask() const
        { return race_mask; }

    void setRaceMask(GPlayerSet mask)
        { race_mask = mask; }

    /** Get specification for parameter.
        \param id which parameter you want to know about
        \return ma_XXX, mat_XXX. Zero if parameter is not required. */
    unsigned getArg(Parameter id) const
        { return arg_spec[id]; }

    void setArg(Parameter id, unsigned spec)
        { arg_spec[id] = spec; }

    /** Get mission flags. \return mf_XXX */
    unsigned getFlags() const
        { return flags; }

    void setFlags(unsigned fl)
        { flags = fl; }

    /** Get mission name. This name should be displayed in selection lists. */
    string_t getName() const
        { return name; }

    void setName(string_t name)
        { this->name = name; }

    string_t getShortName() const
        { return short_name; }

    void setShortName(string_t short_name)
        { this->short_name = short_name; }

    string_t getArgName(Parameter id) const;

    /** Get assigned hot-key. */
    char getHotkey() const
        { return hotkey; }

    /** Assign hot-key. */
    void setHotkey(char c)
        { hotkey = c; }

    /** Set argument name. */
    void setArgName(Parameter id, string_t name)
        { arg_names[id] = name; }

    /** Set condition. */
    void setCondition(string_t cond)
        { exp_condition = cond; }

    /** Set warning condition. */
    void setWarning(string_t warning)
        { exp_warning = warning; }

    /** Set label expression. */
    void setLabel(string_t label)
        { exp_label = label; }

    /** Set "on set" command. */
    void setSetCommand(string_t cmd)
        { cmd_onset = cmd; }

    bool      worksOn(GShip& ship) const;
    string_t  getName(GShip& ship) const;

    /** Mission argument specification. */
    enum {
        ma_TypeMask    = 15,    ///< Type mask.
        ma_Own         = 16,    ///< Own units only.
        ma_NotThis     = 32     ///< This object is not a valid parameter.
    };

    /** Mission argument types (arg_spec & ma_TypeMask). */
    enum {
        mat_Integer    = 1,     ///< Integer.
        mat_PID        = 2,     ///< Planet Id.
        mat_SID        = 3,     ///< Ship Id.
        mat_SID_here   = 4,     ///< Id of a ship here.
        mat_BID        = 5,     ///< Starbase Id.
        mat_Player     = 6      ///< Player number.
    };

    /** Mission flags. */
    enum {
        mf_Waypoint    = 1,     ///< Mission affects waypoint ("Intercept").
        mf_Registered  = 2      ///< Mission is reg-only.
    };

    /** Standard Mission Numbers. */
    enum {
        // Standard missions
        msn_Explore     = 1,
        msn_MineSweep   = 2,
        msn_LayMines    = 3,
        msn_Kill        = 4,
        msn_SensorSweep = 5,
        msn_LnD         = 6,    // Colonize, Land&Disassemble, Decommission
        msn_Tow         = 7,
        msn_Intercept   = 8,
        msn_Special     = 9,
        msn_Cloak       = 10,
        msn_BUFuel      = 11,
        msn_BUDur       = 12,
        msn_BUTri       = 13,
        msn_BUMol       = 14,
        msn_BUSup       = 15,

        // PHost extended missions. These numbers are relative to ExtMissionsStartAt.
        pmsn_BuildTorpsFromCargo = 0,
        pmsn_LayMines            = 1,
        pmsn_LayWeb              = 2,
        pmsn_ScoopTorps          = 3,
        pmsn_GatherBuildTorps    = 4,
        pmsn_BDmc                = 5,
        pmsn_XferTorp            = 6,
        pmsn_XferFtr             = 7,
        pmsn_XferMc              = 8,
        pmsn_StandardSuperSpy    = 9,
        pmsn_Cloak               = 10,
        pmsn_Special             = 11,
        pmsn_GatherBuildFtr      = 12,
        pmsn_BUmc                = 13,
        pmsn_BUclans             = 14,
        pmsn_BeamUpMultiple      = 15,
        pmsn_LayMinesIn          = 16,
        pmsn_LayWebIn            = 17,
        pmsn_Training            = 18
    };
};

/** Mission list. Contains a list of missions. There is a global instance
    as well the possibility to make subset lists. */
class GMissionList {
    typedef std::vector<GMission> container_t;

    container_t data;
    uint32 used_letters;

    static GMissionList instance;
 public:
    typedef container_t::const_iterator iterator;

    GMissionList();
    GMissionList(const GMissionList& other, GShip& ship);
    ~GMissionList();

    /* container interface */
    container_t::size_type getSize() const
        { return data.size(); }
    iterator begin() const
        { return data.begin(); }
    iterator end() const
        { return data.end(); }

    /* manipulator interface */
    bool addMission(const GMission& msn);
    void sort();
    void clear();

    const GMission* getMissionByNumber(int id, GPlayerSet race_mask = GPlayerSet::allUpTo(NUM_PLAYERS)) const;

    void loadFromFile(Stream& in);
    void loadFromIniFile(Stream& in);

    /* accessor interface */
    const GMission& operator[](container_t::size_type n) const
        { return data[n]; }

    static GMissionList& getInstance()
        { return instance; }
};

extern const ModuleInitializer mod_missions[];

bool isMissionCloaking(int mission_id, int owner);

#endif
