/**
  *  \file game/trn-cmd.cc
  *  \brief Applying turn commands to ship/planet
  *
  *  These functions provide the link between the game objects and
  *  GTurnfile. They belong somewhere between save.cc and trn.cc.
  *
  *  General notes: because this code is based on the marshalling
  *  primitives to dissect structures, there's no need to worry about
  *  character sets. getPartialStructure() and storeStructure() will
  *  translate just fine.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/ship.h"
#include "game/planet.h"
#include "game/trn.h"
#include "cpluslib/assert.h"
#include "io/io.h"
#include "game/planet-int.h"
#include "game/ship-int.h"

/* Ships */

/** Apply turn command for a ship. \pre hasShipData() */
void
GShip::applyTurnCommand(GTurnfile& trn, int index)
{
    ASSERT(ship_info);
    getPartialStructure(trn.getCommandData(index),
                        &ship_info->data, ship_info->data.parser,
                        trn.getRecordIndexForCode(trn.getCommandCode(index)),
                        trn.getCommandLength(index));
}

/** Generate turn commands for a ship. \pre hasShipData() */
void
GShip::makeTurnCommands(GTurnfile& trn, const GShip& dis) const
{
    ASSERT(ship_info && dis.ship_info);
    char dat_buffer[TShip::size], dis_buffer[TShip::size];

    storeStructure(dat_buffer, ship_info->data);
    storeStructure(dis_buffer, dis.ship_info->data);
    trn.makeShipCommands(getId(), dis_buffer, dat_buffer);
}

/* Planets */

/** Apply turn commands for a planet.
    \pre hasPlanetData() or hasBaseData(), depending on command. */
void
GPlanet::applyTurnCommand(GTurnfile& trn, int index)
{
    int cmd = trn.getCommandCode(index);
    if (cmd == tcm_PlanetBuildBase) {
        ASSERT(hasPlanetData());
        planet_info->data.build_base = 1;
    } else if (cmd >= tcm_BaseFIRST) {
        ASSERT(hasBaseData());
        getPartialStructure(trn.getCommandData(index),
                            &base_info->data, TStarbase::parser,
                            trn.getRecordIndexForCode(trn.getCommandCode(index)),
                            trn.getCommandLength(index));
    } else {
        ASSERT(hasPlanetData());
        getPartialStructure(trn.getCommandData(index),
                            &planet_info->data, TPlanet::parser,
                            trn.getRecordIndexForCode(trn.getCommandCode(index)),
                            trn.getCommandLength(index));
    }
}

/** Generate turn commands for a planet. \pre hasPlanetData() */
void
GPlanet::makePlanetTurnCommands(GTurnfile& trn, const GPlanet& dis) const
{
    ASSERT(hasPlanetData() && dis.hasPlanetData());
    char dat_buffer[TPlanet::size], dis_buffer[TPlanet::size];

    storeStructure(dat_buffer, planet_info->data);
    storeStructure(dis_buffer, dis.planet_info->data);
    trn.makePlanetCommands(getId(), dis_buffer, dat_buffer);
}

/** Generate turn commands for a starbase. \pre hasBaseData() */
void
GPlanet::makeBaseTurnCommands(GTurnfile& trn, const GPlanet& dis) const
{
    ASSERT(hasBaseData() && dis.hasBaseData());
    char dat_buffer[TStarbase::size], dis_buffer[TStarbase::size];

    storeStructure(dat_buffer, base_info->data);
    storeStructure(dis_buffer, dis.base_info->data);
    trn.makeBaseCommands(getId(), dis_buffer, dat_buffer);
}
