/**
  *  \file int/if/guiif.cc
  *  \brief Interpreter Game Interface - GUI Components
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "int/if/guiif.h"
#include "client/actions.h"
#include "client/cscreen.h"
#include "client/dialogs/keymapdlg.h"
#include "client/dialogs/searchdlg.h"
#include "client/dialogs/selmgr.h"
#include "client/dlg-ammo.h"
#include "client/dlg-build.h"
#include "client/dlg-fcode.h"
#include "client/dlg-mission.h"
#include "client/dlg-structure.h"
#include "client/dlg-supplies.h"
#include "client/dlg-tax.h"
#include "client/dlg-tech.h"
#include "client/dlg-visual.h"
#include "client/screens.h"
#include "client/widgets/specview.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/fcode.h"
#include "game/global.h"
#include "game/host.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/shipquery.h"
#include "int/args.h"
#include "int/compound.h"
#include "int/exec.h"
#include "int/if/ifutil.h"
#include "int/if/userif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "ui/event.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "ui/widgets.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "util/strlist.h"
#include "client/history.h"
#include "game/racename.h"

void
flushUI()
{
    // Tell widgets to order their redraws
    if (haveDisplayedTurn())
        getDisplayedTurn().getCurrentUniverse().doScreenUpdates();

    // Perform redraws
    if (ui_root)
        ui_root->performDeferredRedraws();
}

/** Global command "MessageBox". */
static void
IFMessageBox(IntExecutionContext& exc, IntArgBlock& args)
{
    /* MessageBox <text>[, <heading>] */

    // Validate arguments
    args.checkArgumentCount(1, 2);

    string_t content;
    string_t heading = _("Message");
    if (!checkStringArg(content, args.getNext()))
        return;
    checkStringArg(heading, args.getNext());

    // Do it
    flushUI();
    if (exc.checkForBreak())
        return;
    messageBox(content, heading);
}

/** Global command "UI.Update". */
static void
IFUIUpdate(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    /* UI.Update [<forceFullRedrawFlag>] */

    args.checkArgumentCount(0, 1);

    // FIXME: this is not 100% according to rules, because it treats a missing argument
    // different from a null argument
    int mode = 0;
    if (args.getNumArgs() > 0)
        mode = getBoolValue(args.getNext());

    if (mode >= 0) {
        // Regular redraw
        flushUI();

        // Force full redraw
        if (mode > 0 && ui_root)
            ui_root->drawWidget(true);
    }
}

/** Global command "UI.Input". */
static void
IFUIInput(IntExecutionContext& exc, IntArgBlock& args)
{
    /* UI.Input <prompt>[, <title>, <maxChars>, <flags>, <default>] */
    args.checkArgumentCount(1, 5);

    string_t prompt;
    string_t title;
    string_t defaultText;
    int32_t  maxChars = 255;
    int32_t  flags = 0;
    int32_t  width = ui_root->getExtent().w / 2;

    // Mandatory argument
    if (!checkStringArg(prompt, args.getNext()))
        return;

    // Optional arguments
    title = prompt;
    checkStringArg(title, args.getNext());
    checkIntArg(maxChars, args.getNext(), 0, 32000);
    checkFlagArg(flags, &width, args.getNext(), "NHPFGM");
    checkStringArg(defaultText, args.getNext());

    // Convert flags
    //   N = numeric
    //   H = on high ASCII
    //   P = password masking
    //   F = frame
    //   G = game charset
    //   M = width is in ems
    uint32_t uiFlags = 0;
    if (flags & 1)
        uiFlags |= UIInputLine::ilf_NumbersOnly;
    if (flags & 2)
        uiFlags |= UIInputLine::ilf_NoHi;
    if (flags & 4)
        uiFlags |= UIInputLine::ilf_Hidden;
    if (flags & 8)
        uiFlags |= UIInputLine::ilf_Framed;
    if (flags & 16)
        uiFlags |= UIInputLine::ilf_GameChars;

    if ((flags & 32) == 0 && font_ems[FONT_TITLE] > 0) {
        // width given in in pixels. Convert to ems.
        width /= font_ems[FONT_TITLE];
    }

    // Build a UIInputLine object
    flushUI();
    if (exc.checkForBreak())
        return;
    if (UIInputLine::run(defaultText, maxChars, title, prompt, width, uiFlags)) {
        IntStringValue sv(defaultText);
        exc.setVariable("UI.RESULT", &sv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

static void
IFUIInputNumber(IntExecutionContext& exc, IntArgBlock& args)
{
    // Check arguments
    args.checkArgumentCount(1, 5);

    string_t title;
    if (!checkStringArg(title, args.getNext()))
        return;

    int32_t min = 0;
    checkIntArg(min, args.getNext());

    int32_t max = 10000;
    checkIntArg(max, args.getNext());

    int32_t current = 0;
    checkIntArg(current, args.getNext());

    /* The fifth argument is the help screen Id, which we currently do not use. */

    // Adjust arguments
    if (max < min)
        std::swap(min, max);

    // Do it.
    flushUI();
    if (exc.checkForBreak())
        return;
    UIDecimalSelector nsel(0, min, max, 10);
    nsel.setValue(current);  /* value will be clipped into valid range */
    if (nsel.doStandardDialog(title, title, 0)) {
        IntIntValue iv(nsel.getValue(), false);
        exc.setVariable("UI.RESULT", &iv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

static void
IFUIChooseObject(IntExecutionContext& exc, IntArgBlock& args)
{
    // Check arguments
    int32_t screen;
    args.checkArgumentCount(1);
    if (!checkIntArg(screen, args.getNext(), 0, 100))
        return;

    // Do it
    flushUI();
    if (exc.checkForBreak())
        return;

    int result;
    switch (screen) {
     case 1:
     case 11:
        result = chooseShip();
        break;
     case 2:
     case 12:
        result = choosePlanet();
        break;
     case 3:
     case 13:
        result = chooseBase();
        break;
     // case 6:
     // case 10:
     default:
        throw IntError::rangeError();
    }

    if (result != 0) {
        IntIntValue iv(result, false);
        exc.setVariable("UI.RESULT", &iv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

static void
IFUIMessage(IntExecutionContext& exc, IntArgBlock& args)
{
    using std::tolower;

    args.checkArgumentCount(1, 3);

    // Read arguments
    string_t text;
    string_t title   = _("Message");
    string_t buttons = _("OK");
    if (!checkStringArg(text, args.getNext()))
        return;
    checkStringArg(title, args.getNext());
    checkStringArg(buttons, args.getNext());

    buttons = strTrim(buttons);

    // Prepare dialog
    UIMessageBox msgbox(text, title);
    int id = 0;
    do {
        string_t b = strFirst(buttons, " ");
        if (!b.empty()) {
            ++id;
            msgbox.addButton(id, b, tolower(uint8_t(b[0])));
        }
    } while (strRemove(buttons, " "));
    msgbox.addKey(1, SDLK_RETURN);
    msgbox.addKey(1, ' ');
    msgbox.addKey(id, SDLK_ESCAPE);

    // Do it
    if (id == 0) {
        exc.setVariable("UI.RESULT", 0);
    } else {
        flushUI();
        if (exc.checkForBreak())
            return;
        int result = msgbox.run();
        IntIntValue iv(result, false);
        exc.setVariable("UI.RESULT", &iv);
    }
}

/** Global command "UI.GoToChart" */
static void
IFUIGotoChart(IntExecutionContext& exc, IntArgBlock& args)
{
    // Read arguments
    args.checkArgumentCount(2);
    int x, y;
    if (!checkIntArg(x, args.getNext(), 0, 10000)
        || !checkIntArg(y, args.getNext(), 0, 10000))
        return;

    IntUserInterfaceBinding::getCurrentInstance()->doGoToChart(GPoint(x, y));
    exc.suspendForUI();
}

/** Global command "UI.GoToScreen" */
static void
IFUIGotoScreen(IntExecutionContext& exc, IntArgBlock& args)
{
    // Read arguments
    args.checkArgumentCount(1, 2);
    int32_t screen = 0;
    int32_t obj = 0;
    if (!checkIntArg(screen, args.getNext(), 0, 20))
        return;

    // If second argument is specified but empty, ignore command!
    if (args.getNumArgs() > 0 && !checkIntArg(obj, args.getNext(), 0, 10000))
        return;

    // Validate args
    bool ok;
    switch (screen) {
     case 0:
     case 4:
        // Those don't take args
        ok = (obj == 0);
        break;

     case 1:
     case 11:
        // Ship Id
        if (obj == 0)
            obj = global_ship_type.getCurrent();
        ok = global_ship_type.isExistant(obj);
        break;

     case 2:
     case 12:
        // Planet Id
        if (obj == 0)
            obj = global_planet_type.getCurrent();
        ok = global_planet_type.isExistant(obj);
        break;

     case 3:
     case 13:
        // Base
        if (obj == 0)
            obj = global_base_type.getCurrent();
        ok = global_base_type.isExistant(obj);
        break;

     // case 6:
     // case 10:
     default:
        ok = false;
        break;
    }

    // Valid?
    if (!ok)
        throw IntError::rangeError();

    // do it
    IntUserInterfaceBinding::getCurrentInstance()->doGoToScreen(screen, obj);
    exc.suspendForUI();
}

/** Global command "UI.PopupConsole" */
static void
IFUIPopupConsole(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    IntUserInterfaceBinding::getCurrentInstance()->doPopupConsole();
    exc.suspendForUI();
}

/** Global command "UI.OverlayMessage" */
static void
IFUIOverlayMessage(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    string_t msg;
    if (checkStringArg(msg, args.getNext()) && msg.size()) {
        flushUI();
        IntUserInterfaceBinding::getCurrentInstance()->doOverlayMessage(msg);
        exc.suspendForUI();
    }
}

/** Global command "UI.ListShips" */
static void
IFUIListShips(IntExecutionContext& exc, IntArgBlock& args)
{
    /* UI.ListShips x, y[, flags, ok, heading]
       Flags: F = allow selection of foreign ships; implies A
              A = list all ships
              E = do not display a dialog if there is only one ship
              S = only safe ships, no guessed ones */
    args.checkArgumentCount(2, 5);

    /* Read args */
    int32_t x, y;
    int32_t flags = 0, except = 0;
    string_t ok = _("OK");
    string_t heading = _("List Ships");

    if (!checkIntArg(x, args.getNext(), 0, 10000))
        return;
    if (!checkIntArg(y, args.getNext(), 0, 10000))
        return;
    checkFlagArg(flags, &except, args.getNext(), "FAES");
    checkStringArg(ok, args.getNext());
    checkStringArg(heading, args.getNext());

    /* Validate */
    if (except < 0 || except > NUM_SHIPS)
        throw IntError::rangeError();

    /* Prepare */
    flushUI();
    if (exc.checkForBreak())
        return;

    int dlgFlags = 0;
    if (flags & 1)
        dlgFlags |= ls_IncludeForeignShips | ls_AllowForeignShips;
    if (flags & 2)
        dlgFlags |= ls_IncludeForeignShips;
    if (flags & 4)
        dlgFlags |= ls_EarlyExit;
    if (flags & 8)
        dlgFlags |= ls_SafeShipsOnly;

    /* Do it */
    GObjectId result = doListShips(GPoint(x, y), dlgFlags, except, heading, ok, 0);
    if (result.isValid()) {
        IntIntValue iv(result.getId(), false);
        exc.setVariable("UI.RESULT", &iv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

/** Command 'UI.Search' */
static void
IFUISearch(IntExecutionContext& exc, IntArgBlock& args)
{
    /* UI.Search [text, flags]
       flags: PSBUO = planets, ships, bases, ufos, others
              1234  = name, true, false, location */
    bool immediate = false;
    args.checkArgumentCount(0, 2);
    if (args.getNumArgs() > 0) {
        /* Fetch text */
        immediate = true;
        if (!checkStringArg(search_query.query, args.getNext()))
            return;
    }
    if (args.getNumArgs() > 0) {
        /* Fetch flags */
        int32_t kind = 1;
        int32_t objs = 0;
        CTASSERT(TUserSearchQuery::SearchShips == 1);
        CTASSERT(TUserSearchQuery::SearchPlanets == 2);
        CTASSERT(TUserSearchQuery::SearchBases == 4);
        CTASSERT(TUserSearchQuery::SearchUfos == 8);
        CTASSERT(TUserSearchQuery::SearchOthers == 16);
        if (!checkFlagArg(objs, &kind, args.getNext(), "SPBUO"))
            return;

        switch (kind) {
         case 1:  search_query.search_kind = TUserSearchQuery::MatchName;     break;
         case 2:  search_query.search_kind = TUserSearchQuery::MatchTrue;     break;
         case 3:  search_query.search_kind = TUserSearchQuery::MatchFalse;    break;
         case 4:  search_query.search_kind = TUserSearchQuery::MatchLocation; break;
         default: throw IntError::rangeError();
        }

        search_query.search_obj = objs;
    }

    /* Do it */
    flushUI();
    if (exc.checkForBreak())
        return;

    WScreenId id = doSearch(immediate);
    if (id.getScreenId() == WScreenId::Starchart) {
        IntUserInterfaceBinding::getCurrentInstance()->doGoToChart(id.getPos());
        exc.suspendForUI();
    } else if (id.isValid()) {
        IntUserInterfaceBinding::getCurrentInstance()->doGoToScreen(id.getScreenId(), id.getId());
        exc.suspendForUI();
    } else {
        // nix
    }
}

/** Command 'UI.SelectionManager' */
static void
IFUISelectionManager(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    doSelectionManager();
}

/** Command 'UI.PlanetInfo' */
static void
IFUIPlanetInfo(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int32_t pid;
    args.checkArgumentCount(1);
    if (!checkIntArg(pid, args.getNext(), 1, NUM_PLANETS))
        return;
    flushUI();
    doPlanetScan(pid);
    // exc.suspendForUI();
}

/** Command 'UI.KeymapInfo' */
static void
IFUIKeymapInfo(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    doKeymapDebugger(IntUserInterfaceBinding::getCurrentInstance()->getKeymap());
}

/** Command 'System.ExitRace' */
static void
IFSystemExitRace(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    IntUserInterfaceBinding::getCurrentInstance()->doLeaveRace();
    exc.suspendForUI();
}

/** Internal global command "CC$CSBrowse". */
static void
IFCCCSBrowse(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    /* CC$CSBrowse <forwardFlag>, <onlyMarkedFlag> */
    args.checkArgumentCount(2);

    WControlScreen* cs = dynamic_cast<WControlScreen*>(IntUserInterfaceBinding::getCurrentInstance());
    if (!cs)
        throw IntError::contextError();

    bool forward, marked;
    if (checkBoolArg(forward, args.getNext()) && checkBoolArg(marked, args.getNext()))
        cs->doBrowse(forward, marked);
}

static void
IFCCChangeFCode(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isExistant() || !obj->isPlayed()) {
        throw IntError::contextError();
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(obj)) {
        // Planet fcode change
        GFCodeList list(GFCodeList::getInstance(), *pl);
        WFCodeWindow dialog(_("Change Friendly Code"), list);
        dialog.setFCode(pl->getFCode());
        ui_root->centerWidget(dialog);
        ui_root->add(dialog);
        if (dialog.run() == cm_Enter)
            pl->setFCode(dialog.getFCode());
    } else if (GShip* sh = dynamic_cast<GShip*>(obj)) {
        // Ship fcode change
        GFCodeList list(GFCodeList::getInstance(), *sh);
        WFCodeWindow dialog(_("Change Friendly Code"), list);
        dialog.setFCode(sh->getFCode());
        ui_root->centerWidget(dialog);
        ui_root->add(dialog);
        if (dialog.run() == cm_Enter)
            sh->setFCode(dialog.getFCode());
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangePE(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isExistant() || !obj->isPlayed()) {
        throw IntError::contextError();
    } else if (GShip* sh = dynamic_cast<GShip*>(obj)) {
        // Ship PE change
        StringList d;
        d.add(0, _("0 - none"));
        for (int i = 1; i <= NUM_PLAYERS; ++i)
            d.add(i, format("%X - %s", i, player_racenames.getShortName(i)));

        UIStandardListbox lb(0, -1, 240, d);
        lb.scrollToKey(sh->getPE());
        if (lb.doStandardListbox(_("Change Primary Enemy"), 0))
            sh->setPE(lb.getKey());
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangeMission(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isExistant() || !obj->isPlayed()) {
        throw IntError::contextError();
    } else if (GShip* sh = dynamic_cast<GShip*>(obj)) {
        doShipMissionChange(*sh);
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(obj)) {
        if (pl->hasBase()) {
            doBaseChangeMission(*pl);
        } else {
            throw IntError::contextError();
        }
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCBuildStructures(IntExecutionContext& exc, IntArgBlock& args)
{
    int32_t screen = 0;
    args.checkArgumentCount(1);
    checkIntArg(screen, args.getNext(), 0, 2);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isExistant() || !obj->isPlayed()) {
        throw IntError::contextError();
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(obj)) {
        doPlanetBuildScreen(*pl, screen);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangeTech(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isExistant() && pl->isPlayed() && pl->hasBase()) {
        doTechUpgradeDialog(*pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangeTaxes(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isExistant() && pl->isPlayed()) {
        doTaxation(*pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCSellSupplies(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isExistant() && pl->isPlayed()) {
        WSellSuppliesDialog::doDialog(*pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCBuySupplies(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isExistant() && pl->isPlayed()) {
        WUndoSellSuppliesDialog::doDialog(getDisplayedTurn() /* FIXME */, pl->getId());
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCGive(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isExistant() || !obj->isPlayed()) {
        throw IntError::contextError();
    } else if (!host.isPHost() || !config.CPEnableGive()) {
        // ignore
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(obj)) {
        // give planet
        doGivePlanet(*pl, getDisplayedTurn().getCommands(pl->getOwner()));
    } else if (GShip* sh = dynamic_cast<GShip*>(obj)) {
        // give ship
        doGiveShip(*sh, getDisplayedTurn().getCommands(sh->getOwner()));
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCShipSpec(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject());
    if (sh != 0 && sh->isExistant()) {
        GShipQuery q;
        q.initForExistingShip(getDisplayedTurn().getCurrentUniverse(), sh->getId());
        WSpecView sv;
        sv.setQuery(q);
        sv.doStandardDialog();
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCBuildAmmo(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isExistant() && pl->isPlayed() && pl->hasBase()) {
        doBaseBuildAmmo(*pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCBuildShip(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isExistant() && pl->isPlayed() && pl->hasBase()) {
        WShipBuildScreenWithActor bs(*pl);
        ui_root->centerWidget(bs);
        ui_root->add(bs);
        ui_root->run();
        ui_root->close(bs);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCAddToSim(IntExecutionContext& exc, IntArgBlock& args)
{
    bool ask = true;
    args.checkArgumentCount(0, 1);
    checkBoolArg(ask, args.getNext());
    if (GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject())) {
        if (pl->isExistant() && pl->hasPlanetData()) {
            doPlanetAddToSim(*pl);
        } else {
            throw GError(GError::eNotOwner, _("Nothing known about this planet"));
        }
    } else if (GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject())) {
        flushUI();
        doShipAddToSim(*sh, ask);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCBuildBase(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    if (GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject())) {
        flushUI();
        doPlanetBuildBase(getDisplayedTurn().getCurrentUniverse(), *pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCTransferPlanet(IntExecutionContext& exc, IntArgBlock& args)
{
    bool unload;
    args.checkArgumentCount(1);
    if (!checkBoolArg(unload, args.getNext()))
        return;
    if (GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject())) {
        if (pl->isExistant() && pl->isPlayed()) {
            flushUI();
            doPlanetTransfer(getDisplayedTurn().getCurrentUniverse(), *pl, unload);
        } else {
            throw GError(GError::eNotPlaying, _("You do not play this planet"));
        }
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCTransferShip(IntExecutionContext& exc, IntArgBlock& args)
{
    // Args:
    //   int: kind of transfer (0=ask, 1=ship, 2=planet)
    //   int: target (planet 0: unload)
    int32_t mode, target;
    args.checkArgumentCount(2);
    if (!checkIntArg(mode, args.getNext(), 0, 2) || !checkIntArg(target, args.getNext(), 0, INT_MAX))
        return;

    // Must be our ship
    GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject());
    if (!sh)
        throw IntError::contextError();
    if (!sh->isExistant() || !sh->isPlayed())
        throw GError(GError::eNotPlaying, _("You do not play this ship."));

    // Parse mode/target
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    if (mode == 0) {
        doShipCargoTransfer(univ, *sh);
        return;
    }

    // Parse and validate mode/target
    if (mode == 1) {
        // Ship
        if (!univ.isValidShipId(target) || !univ.getShip(target).isExistant())
            throw GError(GError::eRange, _("This ship does not exist"));
        if (univ.getShip(target).getPos() != sh->getPos())
            throw GError(GError::ePos, _("Different position"));
        if (target == sh->getId())
            throw GError(GError::ePerm, GError::ePerm);
        doShipTransferFor(*sh, GObjectId(global_anyship_type, target));
    } else {
        // Planet
        if (sh->getOrbit() != target)
            throw GError(GError::ePos, _("Different position"));
        if (target == 0) {
            // Jettison
            doShipTransferFor(*sh, GObjectId());
        } else {
            // Planet
            doShipTransferFor(*sh, GObjectId(global_anyplanet_type, target));
        }
    }
}

static void
IFCCPopScreenHistory(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    IntUserInterfaceBinding* b = IntUserInterfaceBinding::getCurrentInstance();
    if (popHistoryScreenId(b->getProperty(iuiScreenRegistered)))
        exc.suspendForUI();
}

static void
IFCCListScreenHistory(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    IntUserInterfaceBinding* b = IntUserInterfaceBinding::getCurrentInstance();
    if (doScreenHistoryDialog(b->getProperty(iuiScreenRegistered)))
        exc.suspendForUI();
}

static void
IFCCViewInbox(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    doInboxScreen();
    exc.suspendForUI();
}

static void
IFCCSendMessage(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    doSendOrReviseMessage();
    exc.suspendForUI();
}

static void
IFChartSetView(IntExecutionContext& exc, IntArgBlock& args)
{
    string_t name;
    args.checkArgumentCount(1);
    if (checkStringArg(name, args.getNext())) {
        IntUserInterfaceBinding::getCurrentInstance()->doSetChartView(strUCase(name));
        exc.suspendForUI();
    }
}

static bool
checkBreak()
{
    bool result = checkBreakEvent();
    if (result) {
        result = UIMessageBox(_("PCC2 is currently executing a script. "
                                "Do you want to abort that script?"),
                              _("Script Interpreter")).doYesNoDialog();
        ui_root->performDeferredRedraws();
    }
    return result;
}

static void
initInterpreterGuiInterface()
{
    // Values
    defineGlobalValue("SYSTEM.GUI", makeBoolValue(1));

    // Procedures
    defineGlobalProcedure("CC$ADDTOSIM",           IFCCAddToSim);
    defineGlobalProcedure("CC$BUILDAMMO",          IFCCBuildAmmo);
    defineGlobalProcedure("CC$BUILDBASE",          IFCCBuildBase);
    defineGlobalProcedure("CC$BUILDSHIP",          IFCCBuildShip);
    defineGlobalProcedure("CC$BUILDSTRUCTURES",    IFCCBuildStructures);
    defineGlobalProcedure("CC$BUYSUPPLIES",        IFCCBuySupplies);
    defineGlobalProcedure("CC$CHANGEFCODE",        IFCCChangeFCode);
    defineGlobalProcedure("CC$CHANGEMISSION",      IFCCChangeMission);
    defineGlobalProcedure("CC$CHANGEPE",           IFCCChangePE);
    defineGlobalProcedure("CC$CHANGETAXES",        IFCCChangeTaxes);
    defineGlobalProcedure("CC$CHANGETECH",         IFCCChangeTech);
    defineGlobalProcedure("CC$CSBROWSE",           IFCCCSBrowse);
    defineGlobalProcedure("CC$GIVE",               IFCCGive);
    defineGlobalProcedure("CC$LISTSCREENHISTORY",  IFCCListScreenHistory);
    defineGlobalProcedure("CC$POPSCREENHISTORY",   IFCCPopScreenHistory);
    defineGlobalProcedure("CC$SELLSUPPLIES",       IFCCSellSupplies);
    defineGlobalProcedure("CC$SENDMESSAGE",        IFCCSendMessage);
    defineGlobalProcedure("CC$SHIPSPEC",           IFCCShipSpec);
    defineGlobalProcedure("CC$TRANSFERPLANET",     IFCCTransferPlanet);
    defineGlobalProcedure("CC$TRANSFERSHIP",       IFCCTransferShip);
    defineGlobalProcedure("CC$VIEWINBOX",          IFCCViewInbox);
    defineGlobalProcedure("CHART.SETVIEW",         IFChartSetView);
    defineGlobalProcedure("MESSAGEBOX",            IFMessageBox);
    defineGlobalProcedure("SYSTEM.EXITRACE",       IFSystemExitRace);
    defineGlobalProcedure("UI.CHOOSEOBJECT",       IFUIChooseObject);
    defineGlobalProcedure("UI.GOTOCHART",          IFUIGotoChart);
    defineGlobalProcedure("UI.GOTOSCREEN",         IFUIGotoScreen);
    defineGlobalProcedure("UI.INPUT",              IFUIInput);
    defineGlobalProcedure("UI.INPUTNUMBER",        IFUIInputNumber);
    defineGlobalProcedure("UI.KEYMAPINFO",         IFUIKeymapInfo);
    defineGlobalProcedure("UI.LISTSHIPS",          IFUIListShips);
    defineGlobalProcedure("UI.MESSAGE",            IFUIMessage);
    defineGlobalProcedure("UI.OVERLAYMESSAGE",     IFUIOverlayMessage);
    defineGlobalProcedure("UI.PLANETINFO",         IFUIPlanetInfo);
    defineGlobalProcedure("UI.POPUPCONSOLE",       IFUIPopupConsole);
    defineGlobalProcedure("UI.SEARCH",             IFUISearch);
    defineGlobalProcedure("UI.SELECTIONMANAGER",   IFUISelectionManager);
    defineGlobalProcedure("UI.UPDATE",             IFUIUpdate);

    IntExecutionContext::setBreakHandler(checkBreak);
}

// Constraints: after mod_intgame_common, after mod_int, with mod_base_gui
const ModuleInitializer mod_intgame_gui[] = {
    { OnStartup, initInterpreterGuiInterface },
    { 0, 0 }
};
