/**
  *  \file int/if/iterif.cc
  *  \brief Interpreter: Iterator Interface
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include "int/if/iterif.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/ion.h"
#include "game/mine.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/ufo.h"
#include "int/args.h"
#include "int/compound.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "int/value.h"

/** Implementation of the "Iterator" function. Iterators implement
    access to a GObjectType. Scripts can use this to access a type's
    current object, and to iterate through the objects. This interface
    is still preliminary.
    - Iterator(n).CurrentIndex: currently active index
    - Iterator(n).Count: number of objects
    - Iterator(n).Id(x): given an index, return that object's Id
    - Iterator(n).Index(id): given an Id, return that object's index; null if none
    - Iterator(n).NextIndex(x,fl): get next index after x. 0 if none found. Flags are "w" to permit wrap, "m" to accept only marked.
    - Iterator(n).PreviousIndex(x,fl): same like NextIndex, but other direction
    - Iterator(n).Object(x): object from index. Still undecided.

    n is:
    - 1,2,3: played ships, planets, bases
    - [10: fleets]
    - 21, 22: all ships, planets
    - 30: ufos
    - 31: ion storms
    - 32: minefields */
IntValue*
IFIteratorGet(IntArgBlock& args)
{
    args.checkArgumentCount(1);
    int32_t v;
    if (!checkIntArg(v, args.getNext()))
        return 0;

    if (GObjectType* ty = getObjectTypeFromIteratorId(v)) {
        return new IntIteratorContext(*ty);
    } else {
        throw IntError::rangeError();
    }
}

GObjectType*
getObjectTypeFromIteratorId(int32_t id)
{
    switch (id) {
     case 1:  return &global_ship_type;
     case 2:  return &global_planet_type;
     case 3:  return &global_base_type;
     case 21: return &global_anyship_type;
     case 22: return &global_anyplanet_type;
     case 30: return &global_ufo_type;
     case 31: return &global_ion_type;
     case 32: return &global_minefield_type;
     default: return 0;
    }
}

namespace {
    enum IntIteratorProperty {
        iitCount,
        iitCurrent,
        iitId,
        iitIndex,
        iitNext,
        iitObject,
        iitPrevious
    };

    class IntIteratorFunction : public IntIndexableValue {
        IntIteratorProperty prop;
        GObjectType& type;
     public:
        IntIteratorFunction(IntIteratorProperty prop, GObjectType& type);

        // IntIndexableValue
        IntValue*   get(IntArgBlock& args);
        void        set(IntArgBlock& args, IntValue* value);
        IntContext* makeFirstContext();

        // IntValue
        string_t    toString(bool readable);
        IntValue*   clone();
        void        store(IntTagNode&, Stream&);
    };
}

static const TIntMapping iter_map[] = {
    { "COUNT",            iitCount,    0 },
    { "CURRENTINDEX",     iitCurrent,  0 },
    { "ID",               iitId,       0 },
    { "INDEX",            iitIndex,    0 },
    { "NEXTINDEX",        iitNext,     0 },
    { "OBJECT",           iitObject,   0 },
    { "PREVIOUSINDEX",    iitPrevious, 0 },
};

IntIteratorFunction::IntIteratorFunction(IntIteratorProperty prop, GObjectType& type)
    : prop(prop), type(type)
{ }

IntValue*
IntIteratorFunction::get(IntArgBlock& args)
{
    int32_t i, fl;
    switch (prop) {
     case iitId:
        // "Id(index)" => get Id of object
        args.checkArgumentCount(1);
        if (checkIntArg(i, args.getNext(), 0, INT_MAX)) {
            if (type.isExistant(i)) {
                return makeIntValue(type.getObject(i).getId());
            } else {
                return 0;
            }
        } else {
            return 0;
        }

     case iitIndex:
        // "Index(id)" => find Id for index
        args.checkArgumentCount(1);
        if (checkIntArg(i, args.getNext(), 0, INT_MAX)) {
            for (int index = type.getNextIndex(0); index != 0; index = type.getNextIndex(index))
                if (type.getObject(index).getId() == i)
                    return makeIntValue(index);
            return 0;
        } else {
            return 0;
        }

     case iitNext:
        // "Next(index[,flags])" => find next index from index
        args.checkArgumentCount(1, 2);
        if (checkIntArg(i, args.getNext(), 0, INT_MAX)) {
            fl = 0;
            checkFlagArg(fl, 0, args.getNext(), "MW");
            if (fl & 2)
                return makeIntValue(type.getNextWrap(i, fl & 1));
            else
                return makeIntValue(type.getNextNoWrap(i, fl & 1));
        } else {
            return 0;
        }

     case iitPrevious:
        // "Previous(index[,flags])" => find previous index from index
        args.checkArgumentCount(1, 2);
        if (checkIntArg(i, args.getNext(), 0, INT_MAX)) {
            fl = 0;
            checkFlagArg(fl, 0, args.getNext(), "MW");
            if (fl & 2)
                return makeIntValue(type.getPreviousWrap(i, fl & 1));
            else
                return makeIntValue(type.getPreviousNoWrap(i, fl & 1));
        } else {
            return 0;
        }

     case iitObject:
        // FIXME
     default:
        return 0;
    }
}

void
IntIteratorFunction::set(IntArgBlock& /*args*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntContext*
IntIteratorFunction::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

// IntValue
string_t
IntIteratorFunction::toString(bool /*readable*/)
{
    return "#<array>";
}

IntValue*
IntIteratorFunction::clone()
{
    return new IntIteratorFunction(prop, type);
}

void
IntIteratorFunction::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}




IntIteratorContext::IntIteratorContext(GObjectType& type)
    : type(type)
{ }

IntIteratorContext::~IntIteratorContext()
{ }

// IntContext:
int32_t
IntIteratorContext::lookup(string_t name)
{
    return lookupName(name, iter_map, countof(iter_map));
}

void
IntIteratorContext::set(int32_t index, IntValue* value)
{
    int32_t v;
    switch (index) {
     case iitCurrent:
        if (checkIntArg(v, value, 0, INT_MAX)) {
            if (type.isExistant(v)) {
                type.setCurrent(v);
            } else {
                throw IntError::rangeError();
            }
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntIteratorContext::get(int32_t index)
{
    switch (IntIteratorProperty(index)) {
     case iitCount:
        {
            int n = 0;
            for (int slot = type.getNextIndex(0); slot != 0; slot = type.getNextIndex(slot))
                ++n;
            return makeIntValue(n);
        }
     case iitCurrent:
        if (type.isExistant(type.getCurrent()))
            return makeIntValue(type.getCurrent());
        else
            return 0;
     case iitId:
     case iitIndex:
     case iitNext:
     case iitObject:
     case iitPrevious:
        return new IntIteratorFunction(IntIteratorProperty(index), type);
     default:
        return 0;
    }
}

IntContext*
IntIteratorContext::clone()
{
    return new IntIteratorContext(type);
}

GObject*
IntIteratorContext::getObject()
{
    return 0;
}

string_t
IntIteratorContext::toString(bool /*readable*/)
{
    // FIXME?
    return "#<iterator>";
}

void
IntIteratorContext::store(IntTagNode& sv, Stream& aux)
{
    // FIXME?
    throw IntError::notSerializable();
}
