%
%  PCC2 Script Self Test
%
%  This is a test for the script interpreter. Originally written for PCC2,
%  it also adapts for PCC 1.x, and thereby highlights the differences
%  between the two implementations. However, it assumes that the
%  improvements made in PCC 1.1.18 are present, and does not adapt to
%  lower versions.
%
%  To run this self-test, open PCC, load a game, and invoke this script.
%  The game can be anything standard, and it will not be modified if
%  everything goes according to plan (hence, if you're not sure, use an
%  unimportant game directory). It is a good idea to run this only on
%  a PCC2 build which passes the t_int_XXX tests :-)
%
%  This tests (almost) all properties and commands provided by PCC's game
%  interface. It tests that they have the correct type, and, if possible,
%  cross-checks them with other properties. It checks that some commands
%  have the expected effect.
%
%  If the test succeeds, it will end with a success message. Otherwise,
%  it will fail with an error.
%
%  Some tests will be omitted when this script thinks they will rightfully
%  fail. In particular, some tests will be omitted when this is run under
%  PCC 1.x. However, since this has to do some cross-checking of data, it
%  will fail if your game has a very funky configuration or the host which
%  produced the RST you're looking at has bugs.
%
%  If a larger group of tests has been omitted, or the group is skipped
%  because of a special property of the game, that will be logged; some
%  smaller groups will be omitted silently.
%

%%%%%%%%%%%%%%%%%%%%%%%% Basic Utilities for Tests %%%%%%%%%%%%%%%%%%%%%%%%

Local IsV2 = (System.Version$ > 199000)
Local GlobalPrefix = If(IsV2, "Global.", "")      % FIXME: can we avoid this?
Local Selftest.NumTests = 0

Print "Self test started for ", System.Program, " ", System.Version

% Check that expression (given as string) is true.
Sub Assert (expr)
  % Note that we use the Eval command, because PCC 1.x has no Eval function.
  Local assert.Result
  Try
    Eval "assert.result := (" & expr & ")"
  Else
    Abort System.Err & " in expression " & expr
  EndTry
  If IsEmpty(Zap(assert.result)) Then
    Abort "Failed assertion: " & expr
  EndIf
  Selftest.NumTests := Selftest.NumTests + 1
EndSub

% Check that two expressions (given as strings) yield the same result
Sub AssertEquals (ax, bx)
  Local ar, br, cr
  Try
    Eval "ar := (" & ax & ")"
  Else
    Abort System.Err & " in expression " & ax
  EndTry
  Try
    Eval "br := (" & bx & ")"
  Else
    Abort System.Err & " in expression " & bx
  EndTry
  If IsEmpty(ar) Then Abort "Failed comparison, expression was null: " & ax
  If IsEmpty(br) Then Abort "Failed comparison, expression was null: " & bx
  Try
    If ar<>br Then Abort "Not equal"
  Else
    Abort "Failed comparison (" & System.Err & "): " & ax & "=" & ar & " doesn't equal " & br & "=" & bx
  EndTry
  Selftest.NumTests := Selftest.NumTests + 1
EndSub

% Check that an expression (given as string) yields null.
Sub AssertNull (expr)
  Local rr
  Try
    Eval "rr := (" & expr & ")"
  Else
    Abort System.Err & " in expression " & expr
  EndTry
  If Not IsEmpty(rr) Then Abort "Failed assertion, expression not null: " & expr & "=" & rr
  Selftest.NumTests := Selftest.NumTests + 1
EndSub

% Check that a statement/expression throws
Sub AssertThrows (stmt)
  Try
    Eval stmt
  Else
    Selftest.NumTests := Selftest.NumTests + 1
    Return
  EndTry
  Abort "Statement did not throw: " & stmt
EndSub


%%%%%%% Global Variables (before possible environment modifications) %%%%%%

Assert "a:=1"
Assert "b:=1"
Assert "c:=1"
Assert "d:=1"
Assert "e:=1"
Assert "f:=1"
Assert "g:=1"
Assert "h:=1"
Assert "i:=1"
Assert "j:=1"
Assert "k:=1"
Assert "l:=1"
Assert "m:=1"
Assert "n:=1"
Assert "o:=1"
Assert "p:=1"
Assert "q:=1"
Assert "r:=1"
Assert "s:=1"
Assert "t:=1"
Assert "u:=1"
Assert "v:=1"
Assert "w:=1"
Assert "x:=1"
Assert "y:=1"
Assert "UI.Result:=1"
If IsV2 Then Assert "System.Err:='a'"        % property in V1, (localizable, assignable) variable in V2
Assert "Build.Remainder:=1"
Assert "Cargo.Remainder:='1N'"
Assert "UI.Prefix:=1"

% Test 'Global.' prefix. This works in 1.1.18 and in PCC2.
AssertEquals "Global.UI.Prefix", "1"
Shared Global.gv0
Assert "gv0 := 2"
AssertEquals "gv0", "2"
AssertEquals "Global.gv0", "2"
Assert "Global.gv0 := 7"
AssertEquals "gv0", "7"
AssertEquals "Global.gv0", "7"


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Command Tests %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
%  Dim (should be first because hell breaks loose if it is wrong)
%

% Localize a. Check correct compilation (the compiler must not optimize by
% making 'a' predefined when it is used before localisation).
Sub d0.test
  If IsEmpty(Zap(a=1)) Then Abort "Failed localisation #1"
  Local a
  If Not IsEmpty(a) Then Abort "Failed localisation #2"
  a:=3
  AssertEquals "a", "3"

  % Redefining doesn't change its value
  Local a:=4
  AssertEquals "a", "3"

  % Redefining a shared variable doesn't change its value...
  b:=5
  AssertEquals "b", "5"
  Shared b:=6
  AssertEquals "b", "5"

  % ...but localizing it does.
  Local b:=7
  AssertEquals "b", "7"

  % ...but localizing it again doesn't.
  Local b:=8
  AssertEquals "b", "7"
  Dim b:=9
  AssertEquals "b", "7"
EndSub
d0.test

b:=7
If IsV2 Then
  % Localizing using a self-reference uses the previous value (use a subroutine
  % so we can initialize 'b' without mentioning it in the test's code).
  % This does not hold in v1.x, which defines the variable before the initializer.
  Sub d1.test
    AssertEquals "b", "7"
    Dim b:=b+1
    AssertEquals "b", "8"
  EndSub
  d1.test
EndIf

% This shows how a variable can be accessed by a subroutine without being
% mentioned in it.
a := 99
Sub d2.access1
  b := a
EndSub
Sub d2.access2
  d2.access1
  Local a
EndSub
d2.access2
AssertEquals "b", "99"

% Same thing, using Eval
a := 88
Sub d3.access (which)
  Eval "b := " & which
  Local a
EndSub
d3.access "a"
AssertEquals "b", "88"

% Same thing, using RunHook
a := 77
On d4.access Do b := a
Sub d4.access
  RunHook d4.access
  Local a
EndSub
d4.access
AssertEquals "b", "77"

% Localizing must still work, of course
a := 99
Sub d5.access1
  b := a
EndSub
Sub d5.access2
  Local a
  d5.access1
EndSub
d5.access2
AssertNull "b"


%
%  Try, TryLoad, and Abort
%
Local testvar := 1
Try
  Abort 'ha'
  testvar := testvar + 1
Else
  Assert 'System.Err <> ""'
  testvar := testvar + 2
EndTry
AssertEquals "testvar", "3"


Try
  testvar := 2/0
  testvar := 99
EndTry
AssertEquals "testvar", "3"


Try
  testvar := 0
  TryLoad "/some/nonexistant/file"     % a file "guaranteed" to not exist
  testvar := 10
Else
  testvar := 20
EndTry
AssertEquals "testvar", "10"


Try
  testvar := 0
  TryLoad "gen" & My.Race$ & ".dat"    % a file "guaranteed" to exist but not contain a valid script
  TryLoad "player" & My.Race$ & ".rst" % ...and if the above one didn't exist, this one should
  testvar := 10
Else
  testvar := 20
EndTry
AssertEquals "testvar", "20"

Try Abort "foo"
AssertEquals "System.Err", "'foo'"


%
%  Functions (V2 only)
%
If IsV2 Then
  % Test case for stack handling: "zz()+zz()" must return 8.
  % This found a bug in 'Local' and one in 'evals' so far.
  Sub t1.test
    Local i
    i := 3
  EndSub

  Function t1.zz
    t1.test
    t1.sel 1
    t1.sel 2
    t1.sel 3
    t1.sel 4
    t1.sel 5
    t1.sel 6
    t1.sel 20
    t1.sel 90
    t1.sel 101
    Return 4
  EndFunction

  Sub t1.Print(X)
    % dummy; mainly to avoid that t1.sel is optimized away
  EndSub

  Sub t1.sel(z)
    Select Case z
      Case 1
        t1.Print "one"
      Case 2
        t1.Print "two"
      Case 3,4,5
        t1.Print "three to five"
      Case Is < 10, Is > 100
        t1.Print "small or huge"
      Case Else
        t1.Print "middle"
    EndSelect
  EndSub

  AssertEquals "t1.zz() + t1.zz()", "8"

  % Recursion
  Function t2.fib(n)
    If n < 2
      Return 1
    Else
      Return t2.fib(n-1) + t2.fib(n-2)
    Endif
  EndFunction

  AssertEquals "t2.fib(1)", "1"
  AssertEquals "t2.fib(2)", "2"
  AssertEquals "t2.fib(3)", "3"
  AssertEquals "t2.fib(4)", "5"
  AssertEquals "t2.fib(5)", "8"
  AssertEquals "t2.fib(6)", "13"
  AssertEquals "t2.fib(7)", "21"
  AssertEquals "t2.fib(8)", "34"
  AssertEquals "t2.fib(9)", "55"
  AssertEquals "t2.fib(10)", "89"
Else
  Print "Not checking functions in PCC 1.x"
EndIf


%
%  Subroutines
%

Sub s0.test (a, b, Optional c)
  If a<>1 Then Abort "a failed"
  If b<>2 Then Abort "b failed"
  If Not IsEmpty(c) Then Abort "c failed"
EndSub
s0.test 1,2
s0.test 1,2,Z(0)
a := Z(0)
s0.test 1,2,a


% Modified example from PCC1 manual
a := ''
Sub s1.Hack
  Sub s1.Hack
    a := a & "foo"
  EndSub
  s1.Hack
  a := a & "bar"
EndSub
s1.Hack
AssertEquals "a", "'foobar'"


%
%  Eval
%
If IsV2 Then
  Sub e1.proc (expr)
    Local a = Eval(expr)
  EndSub

  Function e1.func
    e1.proc "1+1"
    Return 1
  EndFunction

  AssertEquals "e1.func() & e1.func()", "'11'"
Else
  Print "Not checking 'Eval()' function in PCC 1.x"
EndIf

AssertThrows "Eval '1+1)'"


% Test nested Eval, including those with mixed contexts (one/multiline)
Eval "a:=1"
AssertEquals "a", "1"
Eval "Eval 'a:=2'"
AssertEquals "a", "2"
Eval "For i:=1 to 10 do Eval 'a:=a+' & i"
AssertEquals "a", "57"
Eval "For i:=1 to 10", "Eval 'a:=a+' & i", "Next"
AssertEquals "a", "112"
Eval "For i:=1 to 10", "Eval 'local a', 'a:=' & i", "Next"
AssertEquals "a", "112"
If a>100 Then Eval "For i:=1 to 10", "Eval 'a:=a+' & i", "Next"
AssertEquals "a", "167"


%
%  Arrays (V2 only)
%
If IsV2 Then
  Sub array1
    Local a(100)
    Local i
    For i:=0 To 99 Do AssertNull "a(i)"
    For i:=0 To 99 Do a(i):=i*i
    For i:=0 To 99 Do AssertEquals "a(i)", "i*i"
    AssertThrows "a(100)"
    AssertThrows "a(1,2)"
    AssertThrows "a()"
  EndSub
  Sub array2
    Local a(100,50)
    Local i, j
    For i:=0 To 99 Do For j:=0 To 49 Do AssertNull "a(i,j)"
    For i:=0 To 99 Do For j:=0 To 49 Do a(i,j):=i*j
    For i:=0 To 99 Do For j:=0 To 49 Do AssertEquals "a(i,j)", "i*j"
    AssertThrows "a(100,1)"
    AssertThrows "a(1,50)"
    AssertThrows "a(1)"
    AssertThrows "a(1,5)"
    AssertThrows "a()"
  EndSub
EndIf


%
%  On/RunHook
%

Dim Shared Selftest.DidRunHookTest
If Selftest.DidRunHookTest Then
  % Because hooks cannot be cleared, we can only do this once per session;
  % otherwise, the remains from last time will bite us.
  Print "Not checking 'On'/'RunHook', we already did that this session."
Else
  Selftest.DidRunHookTest := 1

  % Basic behaviour
  a:=0
  AssertEquals "a", "0"
  On SomeTest Do a:=1
  AssertEquals "a", "0"
  RunHook SomeTest
  AssertEquals "a", "1"

  % Adding the same thing several times should not optimize one away.
  % Also, this found a bug in the compilation of "On".
  For i:=1 To 5 Do On SomeTest Do a:=a+2
  AssertEquals "a", "1"
  RunHook SomeTest
  AssertEquals "a", "11"

  % A hook adding to itself may or may not execute the just-added part,
  % but shouldn't crash in any case
  On SomeTest Do On SomeTest Do a:=a+3
  RunHook SomeTest
  Assert "a=11 or a=14"

  RunHook SomeTest
  Assert "a=14 or a=17"
EndIf


%
%  Nested Loops and Conditionals
%
T:=10
For i:=1 To 10 Do
  j := i
  If i=5 Then Break
Next
AssertEquals "j", 5

For i:=1 To 10 Do
  Select Case i
    Case 1,3,5,7,9
      j:=i
    Case Is >6
      Break
    Case 4
      k:=i
  EndSelect
Next
AssertEquals "j", "7"
AssertEquals "k", "4"

i:=0
Do While i<10
  i:=i+1
  Select Case i
    Case 3
      Continue
    Case 4
      Break
  EndSelect
  j:=i
Loop
AssertEquals "j", "2"
AssertEquals "i", "4"

% The For loops again with a variable upper bound to defeat the optimizer
For i:=1 To T Do
  j := i
  If i=5 Then Break
Next
AssertEquals "j", 5

For i:=1 To T Do
  Select Case i
    Case 1,3,5,7,9
      j:=i
    Case Is >6
      Break
    Case 4
      k:=i
  EndSelect
Next
AssertEquals "j", "7"
AssertEquals "k", "4"


% Select-case with type mismatch is a runtime error
Try
  j:=0
  Select Case i
    Case "a"
      j:=1
    Case "b"
      j:=1
    Case Else
      j:=1
  EndSelect
  j:=1
EndTry
AssertEquals "j", "0"

% Although this is legal in v1, it is interpreted wrong. 1.x attempts to
% associate the 'Else' with the 'Try', which fails or course. This is actually
% a (very hard to fix) bug in PCC 1.x.
If IsV2 Then
  i:=1
  Try
    If 1
      Abort "foo"
      i:=9
    Else
      j:=10
    EndIf
  EndTry
  AssertEquals "i", "1"
EndIf


%
%  Comparisons
%
Sub testCompare
  If Not ("a" < "[") Then Abort "Error compare 1"
  If Not ("A" < "[") Then Abort "Error compare 1a"
  If Not ("[" > "a") Then Abort "Error compare 2"
  If Not ("[" > "A") Then Abort "Error compare 2a"
  If Not ("a" = "A") Then Abort "Error compare 3"
  If Not ("a" <> "") Then Abort "Error compare 4"
EndSub
testCompare

%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Cargo Functions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% CAdd
AssertEquals "CAdd('100T', '50T')", "'150T'"
AssertEquals "CAdd('100T', '100D')", "'100T 100D'"
AssertEquals "CAdd('100$', '100S')", "'100S 100$'"
AssertEquals "CAdd('100$ 50$ 20$', '100S')", "'100S 170$'"

% CSub
AssertEquals "CSub('100T', '50T')", "'50T'"
AssertEquals "CSub('100S', '50$')", "'50S'"
AssertEquals "CSub('100T', '50T', '20T')", "'30T'"

% CExtract
AssertEquals "CExtract('100T 200D 50M', 'TD')", "300"
AssertEquals "CExtract('100T 200D 50M', 'TDMTDM')", "350"

AssertEquals "CExtract('100T 200D 300M 400N 500S 600$ 700F 800W', 'T')", "100"
AssertEquals "CExtract('100T 200D 300M 400N 500S 600$ 700F 800W', 'D')", "200"
AssertEquals "CExtract('100T 200D 300M 400N 500S 600$ 700F 800W', 'M')", "300"
AssertEquals "CExtract('100T 200D 300M 400N 500S 600$ 700F 800W', 'N')", "400"
AssertEquals "CExtract('100T 200D 300M 400N 500S 600$ 700F 800W', 'S')", "500"
AssertEquals "CExtract('100T 200D 300M 400N 500S 600$ 700F 800W', '$')", "600"
AssertEquals "CExtract('100T 200D 300M 400N 500S 600$ 700F 800W', 'F')", "700"
AssertEquals "CExtract('100T 200D 300M 400N 500S 600$ 700F 800W', 'W')", "800"

% CRemove
AssertEquals "CRemove('200T 100D', 'T')", "'100D'"

% CMul
AssertEquals "CMul('10T 20D', 15)", "'150T 300D'"
a := CMul('1TDM 42$', 10)
Assert "a='10TDM 420$' or a='10T 10D 10M 420$'"

AssertNull "CMul(Z(0), '1T')"
AssertNull "CMul(Z(0), Z(0))"
AssertNull "CMul('1T', Z(0))"

% CDiv
AssertEquals "CDiv('100T 30M 50S', '3T 2M 5S')", "10"
AssertEquals "CDiv('100T 30M 150S', '3T 2M 5S')", "15"
AssertEquals "CDiv('10T 30M 150S', '3T 2M 5S')", "3"
AssertEquals "CDiv('10T 30M 150S', 3)", "'3T 10M 50S'"

AssertEquals "CDiv('100T 75S 25$', '1T 1$')", "100"
AssertEquals "CDiv('100T 75S 25$', '2T 1$ 1S')", "50"
AssertEquals "CDiv('100T 75S 25$', '3T 1$ 1S')", "33"
AssertEquals "CDiv('100T 75S 25$', '1T 3$ 7S')", "10"

AssertNull "CDiv(Z(0), '1T')"
AssertNull "CDiv(Z(0), Z(0))"
AssertNull "CDiv('1T', Z(0))"

% CCompare
Assert "CCompare('1T', '')"
Assert "not CCompare('', '1T')"
Assert "CCompare('', '')"
Assert "CCompare('1T', '1T')"
Assert "CCompare('100$ 100S', '200$')"
Assert "not CCompare('200$', '100$ 100S')"

AssertNull "CCompare(Z(0), '1T')"
AssertNull "CCompare(Z(0), Z(0))"
AssertNull "CCompare('1T', Z(0))"

If IsV2 Then
  % v1.x fails these
  AssertNull "CAdd(Z(0), '1T')"
  AssertNull "CAdd(Z(0), Z(0))"
  AssertNull "CAdd('1T', Z(0))"

  AssertNull "CSub(Z(0), '1T')"
  AssertNull "CSub(Z(0), Z(0))"
  AssertNull "CSub('1T', Z(0))"

  AssertNull "CExtract(Z(0), '1T')"
  AssertNull "CExtract(Z(0), Z(0))"
  AssertNull "CExtract('1T', Z(0))"

  AssertNull "CRemove(Z(0), '1T')"
  AssertNull "CRemove(Z(0), Z(0))"
  AssertNull "CRemove('1T', Z(0))"
Else
  Print "Not checking null 'CAdd' etc. in PCC 1.x"
EndIf

%%%%%%%%%%%%%%%%%%%%%%%%%%%% Global Properties %%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
%  Globals
%
Assert "System.GameType='Registered' Or System.GameType='Shareware'"
Assert "System.GameType$=0 Or System.GameType$=1"
Assert "System.Err Or True"
Assert "System.GUI=True or System.GUI=False"
Assert "System.Host='Host' or System.Host='SRace' or System.Host='PHost'"
Assert "System.Host$=0 or System.Host$=1 or System.Host$=2"
Assert "IsNum(System.HostVersion)"
Assert "IsString(System.Program)"
Assert "IsString(System.RegStr1)"
Assert "IsString(System.RegStr2)"
Assert "IsString(System.Version)"
Assert "IsNum(System.Version$)"
Assert "IsNum(Turn)"
Assert "IsString(Turn.Date)"
Assert "IsString(Turn.Time)"
Assert "IsNum(System.RandomSeed)"

AssertEquals "System.GameType", "If(System.GameType$, 'Shareware', 'Registered')"
Select Case System.Host$
  Case 0
    AssertEquals "System.Host", "'Host'"
  Case 1
    AssertEquals "System.Host", "'SRace'"
  Case 2
    AssertEquals "System.Host", "'PHost'"
EndSelect


%
%  My Race
%
Assert "IsNum(My.Bases)   And My.Bases >= 0   And My.Bases <= 500"
Assert "IsNum(My.Planets) And My.Planets >= 0 And My.Planets <= 500"
Assert "IsString(My.Race)"
Assert "IsNum(My.Race$) And My.Race$ > 0 And My.Race$ <= 11"
Assert "IsString(My.Race.Adj)"
Assert "IsString(My.Race.Full)"
Assert "IsNum(My.Race.Id) And My.Race.Id > 0 And My.Race.Id <= 12"
Assert "IsNum(My.Race.Mission) And My.Race.Mission > 0 And My.Race.Mission <= 12"
Assert "IsNum(My.Score) And My.Score >= 0"
Assert "IsNum(My.Ships)            And My.Ships >= 0            And My.Ships <= 999"
Assert "IsNum(My.Ships.Capital)    And My.Ships.Capital >= 0    And My.Ships.Capital <= 999"
Assert "IsNum(My.Ships.Freighters) And My.Ships.Freighters >= 0 And My.Ships.Freighters <= 999"
Assert "IsNum(My.Team) And My.Team >= 0"

AssertEquals "Player(My.Race$).Race", "My.Race.Full"          % FIXME: inconsistent naming
AssertEquals "Player(My.Race$).Race.Short", "My.Race"         % FIXME: inconsistent naming
AssertEquals "Player(My.Race$).Race.Adj", "My.Race.Adj"
AssertEquals "My.Ships", "My.Ships.Capital + My.Ships.Freighters"
AssertEquals "My.Score", "120*My.Bases + 10*My.Planets + 10*My.Ships.Capital + My.Ships.Freighters"

If Cfg("CPEnableRemote") Then
  Print "Not checking scores vs. unit counts, CPEnableRemote is active."
Else
  AssertEquals "My.Planets",          "CountPlanets(Owner$=My.Race$)"
  AssertEquals "My.Bases",            "CountPlanets(Owner$=My.Race$ And Base.YesNo)"
  AssertEquals "My.Ships",            "CountShips(Owner$=My.Race$)"
  AssertEquals "My.Ships.Freighters", "CountShips(Owner$=My.Race$ And Not (Fighter.Bays Or Beam.Count Or Torp.LCount))"
  AssertEquals "My.Ships.Capital",    "CountShips(Owner$=My.Race$ And (Fighter.Bays Or Beam.Count Or Torp.LCount))"
EndIf


%
%  Players
%
For i:=1 To 11 Do
  Assert "IsString(Player(i).Race)"
  Assert "IsString(Player(i).Race.Short)"
  Assert "IsString(Player(i).Race.Adj)"
  Assert "IsNum(Player(i).Bases) or IsEmpty(Player(i).Bases)"
  Assert "IsNum(Player(i).Planets) or IsEmpty(Player(i).Planets)"
  AssertEquals "Player(i).Race$", "i"
  Assert "IsNum(Player(i).Race.Id)"
  Assert "IsNum(Player(i).Race.Mission)"
  assert "IsNum(Player(i).Team)"
  If Not IsEmpty(Player(i).Score) Then
    AssertEquals "Player(i).Score", "120*Player(i).Bases + 10*(Player(i).Ships.Capital + Player(i).Planets) + Player(i).Ships.Freighters"
    AssertEquals "Player(i).Ships", "Player(i).Ships.Capital + Player(i).Ships.Freighters"
  EndIf
Next

For i:=1 To 11 Do
  With Player(i) do
    Assert "IsString(Race)"
    Assert "IsString(Race.Short)"
    Assert "IsString(Race.Adj)"
    Assert "IsNum(Bases) or IsEmpty(Bases)"
    Assert "IsNum(Planets) or IsEmpty(Planets)"
    Assert "IsNum(Race.Id)"
    Assert "IsNum(Race.Mission)"
    Assert "IsNum(Team)"
    If Not IsEmpty(Score) Then
      AssertEquals "Score", "120*Bases + 10*(Ships.Capital + Planets) + Ships.Freighters"
      AssertEquals "Ships", "Ships.Capital + Ships.Freighters"
    EndIf
  EndWith
Next

ForEach Player
  Assert "IsString(Race)"
  Assert "IsString(Race.Short)"
  Assert "IsString(Race.Adj)"
  Assert "IsNum(Bases) or IsEmpty(Bases)"
  Assert "IsNum(Planets) or IsEmpty(Planets)"
  Assert "IsNum(Race.Id)"
  Assert "IsNum(Race.Mission)"
  Assert "IsNum(Team)"
  If Not IsEmpty(Score) Then
    AssertEquals "Score", "120*Bases + 10*(Ships.Capital + Planets) + Ships.Freighters"
    AssertEquals "Ships", "Ships.Capital + Ships.Freighters"
  EndIf
Next
ForEach Global.Player
  Assert "IsString(Race)"
Next

If IsV2 Then
  b := Player(3).Race
  a := Find(Player, Race = b, Race$)
  Assert "a > 0"
  Assert "a <= 3"

  a := Count(Player)
  AssertEquals "a", "11"

  a := Count(Player, Race = b)
  Assert "a >= 1"
Else
  Print "Not checking 'Find/Count Player' in PCC 1.x"
EndIf


%
%  Beams
%
For i:=1 To 10 Do
  Assert "IsNum(Beam(i).Cost.T)"
  Assert "IsNum(Beam(i).Cost.D)"
  Assert "IsNum(Beam(i).Cost.M)"
  Assert "IsNum(Beam(i).Cost.MC)"
  Assert "IsString(Beam(i).Cost.Str)"
  AssertEquals "CExtract(Beam(i).Cost.Str, 'T')", "Beam(i).Cost.T"
  AssertEquals "CExtract(Beam(i).Cost.Str, 'D')", "Beam(i).Cost.D"
  AssertEquals "CExtract(Beam(i).Cost.Str, 'M')", "Beam(i).Cost.M"
  AssertEquals "CExtract(Beam(i).Cost.Str, '$')", "Beam(i).Cost.MC"
  Assert "IsNum(Beam(i).Damage)"
  Assert "IsNum(Beam(i).Kill)"
  Assert "IsNum(Beam(i).Mass)"
  Assert "IsNum(Beam(i).Tech)"
  Assert "IsNum(Beam(i).Tech.Beam)"
  Assert "IsString(Beam(i).Name)"
  Assert "IsString(Beam(i).Name.Short)"

  AssertEquals "Beam(i).Tech", "Beam(i).Tech.Beam"
  Assert "Beam(i).Tech > 0 And Beam(i).Tech <= 10"
Next

% Test that beams are With-able and ForEach-able.
With Beam(2) Do
  AssertEquals "Tech", "Tech.Beam"
  Assert "IsNum(Tech)"
EndWith
ForEach Beam
  AssertEquals "Tech", "Tech.Beam"
  Assert "IsNum(Tech)"
Next
ForEach Global.Beam Do Assert "IsNum(Tech)"
If IsV2
  AssertEquals "Count(Beam)", 10
EndIf


%
%  Torpedoes
%
For i:=1 To 10 Do
  % Torpedo()
  Assert "IsNum(Torpedo(i).Cost.T)"
  Assert "IsNum(Torpedo(i).Cost.D)"
  Assert "IsNum(Torpedo(i).Cost.M)"
  Assert "IsNum(Torpedo(i).Cost.MC)"
  Assert "IsString(Torpedo(i).Cost.Str)"
  AssertEquals "CExtract(Torpedo(i).Cost.Str, 'T')", "Torpedo(i).Cost.T"
  AssertEquals "CExtract(Torpedo(i).Cost.Str, 'D')", "Torpedo(i).Cost.D"
  AssertEquals "CExtract(Torpedo(i).Cost.Str, 'M')", "Torpedo(i).Cost.M"
  AssertEquals "CExtract(Torpedo(i).Cost.Str, '$')", "Torpedo(i).Cost.MC"
  Assert "IsNum(Torpedo(i).Damage)"
  Assert "IsNum(Torpedo(i).Kill)"
  Assert "IsNum(Torpedo(i).Mass)"
  Assert "IsNum(Torpedo(i).Tech)"
  Assert "IsNum(Torpedo(i).Tech.Torpedo)"
  Assert "IsString(Torpedo(i).Name)"
  Assert "IsString(Torpedo(i).Name.Short)"

  AssertEquals "Torpedo(i).Tech", "Torpedo(i).Tech.Torpedo"
  Assert "Torpedo(i).Tech > 0 And Torpedo(i).Tech <= 10"

  % Launcher()
  Assert "IsNum(Launcher(i).Cost.T)"
  Assert "IsNum(Launcher(i).Cost.D)"
  Assert "IsNum(Launcher(i).Cost.M)"
  Assert "IsNum(Launcher(i).Cost.MC)"
  Assert "IsString(Launcher(i).Cost.Str)"
  AssertEquals "CExtract(Launcher(i).Cost.Str, 'T')", "Launcher(i).Cost.T"
  AssertEquals "CExtract(Launcher(i).Cost.Str, 'D')", "Launcher(i).Cost.D"
  AssertEquals "CExtract(Launcher(i).Cost.Str, 'M')", "Launcher(i).Cost.M"
  AssertEquals "CExtract(Launcher(i).Cost.Str, '$')", "Launcher(i).Cost.MC"
  Assert "IsNum(Launcher(i).Damage)"
  Assert "IsNum(Launcher(i).Kill)"
  Assert "IsNum(Launcher(i).Mass)"
  Assert "IsNum(Launcher(i).Tech)"
  Assert "IsNum(Launcher(i).Tech.Torpedo)"
  Assert "IsString(Launcher(i).Name)"
  Assert "IsString(Launcher(i).Name.Short)"

  % Relations
  AssertEquals "Torpedo(i).Damage", "Launcher(i).Damage"
  AssertEquals "Torpedo(i).Kill", "Launcher(i).Kill"
  AssertEquals "Torpedo(i).Tech", "Launcher(i).Tech"
  AssertEquals "Torpedo(i).Mass", "1"
  AssertEquals "Torpedo(i).Cost.T", "1"
  AssertEquals "Torpedo(i).Cost.D", "1"
  AssertEquals "Torpedo(i).Cost.M", "1"
Next

% Test that torpedoes are With-able and ForEach-able.
With Torpedo(2) Do
  AssertEquals "Tech", "Tech.Torpedo"
  Assert "IsNum(Tech)"
EndWith
ForEach Torpedo
  AssertEquals "Tech", "Tech.Torpedo"
  Assert "IsNum(Tech)"
Next
ForEach Global.Torpedo Do Assert "IsNum(Tech)"
If IsV2
  AssertEquals "Count(Torpedo)", 10
EndIf

% Test that launchers are With-able and ForEach-able.
With Launcher(2) Do
  AssertEquals "Tech", "Tech.Torpedo"
  Assert "IsNum(Tech)"
EndWith
ForEach Launcher
  AssertEquals "Tech", "Tech.Torpedo"
  Assert "IsNum(Tech)"
Next
ForEach Global.Launcher Do Assert "IsNum(Tech)"
If IsV2 Then
  AssertEquals "Count(Launcher)", 10
EndIf


%
%  Hulls
%
For i:=1 To 50 Do       % PCC allows for any number of hulls but I guess every game will have at least 50
  Assert "IsNum(Hull(i).Beam.Max)"
  Assert "IsNum(Hull(i).Cargo.Max)"
  Assert "IsNum(Hull(i).Cargo.MaxFuel)"
  Assert "IsNum(Hull(i).Cost.T)"
  Assert "IsNum(Hull(i).Cost.D)"
  Assert "IsNum(Hull(i).Cost.M)"
  Assert "IsNum(Hull(i).Cost.MC)"
  Assert "IsString(Hull(i).Cost.Str)"
  Assert "IsNum(Hull(i).Crew.Normal)"
  Assert "IsNum(Hull(i).Engine.Count)"
  Assert "IsNum(Hull(i).Fighter.Bays)"
  Assert "IsNum(Hull(i).Image)"
  Assert "IsNum(Hull(i).Image$)"
  Assert "IsNum(Hull(i).Mass)"
  Assert "IsString(Hull(i).Name)"
  Assert "IsString(Hull(i).Name.Short)"
  Assert "IsString(Hull(i).Special)"
  Assert "IsNum(Hull(i).Tech)"
  Assert "IsNum(Hull(i).Tech.Hull)"
  Assert "IsNum(Hull(i).Torp.LMax)"

  % Hull().Id exists since 1.1.18.
  Assert "IsNum(Hull(i).Id)"
  AssertEquals "Hull(i).Id", "i"

  AssertEquals "Hull(i).Tech", "Hull(i).Tech.Hull"
  AssertEquals "CExtract(Hull(i).Cost.Str, 'T')", "Hull(i).Cost.T"
  AssertEquals "CExtract(Hull(i).Cost.Str, 'D')", "Hull(i).Cost.D"
  AssertEquals "CExtract(Hull(i).Cost.Str, 'M')", "Hull(i).Cost.M"
  AssertEquals "CExtract(Hull(i).Cost.Str, '$')", "Hull(i).Cost.MC"
  Assert "Hull(i).Torp.LMax=0 or Hull(i).Fighter.Bays=0"
Next

% Test that hulls are With-able and ForEach-able.
With Hull(2) Do
  AssertEquals "Tech", "Tech.Hull"
  Assert "IsNum(Tech)"
EndWith
ForEach Hull
  AssertEquals "Tech", "Tech.Hull"
  Assert "IsNum(Tech)"
Next
ForEach Global.Hull Do Assert "IsNum(Tech)"
If IsV2
  Assert "Count(Hull) >= 50"
EndIf

% Figure out how many hulls we have. This is used later on
Local Selftest.NumHulls = 0
If IsV2 Then
  % v2: indexing Hull() out of range gets a range error
  Try
    Do
      Hull(Selftest.NumHulls+1).Name
      Selftest.NumHulls := Selftest.NumHulls + 1
    Loop
  EndTry
Else
  % v1: indexing Hull() out of range yields null
  Do While Hull(Selftest.NumHulls+1).Name
    Selftest.NumHulls := Selftest.NumHulls + 1
  Loop
EndIf

% Check TrueHull
For p:=1 To 11
  For i:=1 To 20
    Assert "IsNum(Truehull(i,p))"
    Assert "Truehull(i,p)>=0 and Truehull(i,p)<="&Selftest.NumHulls
  Next
  AssertNull "Truehull(z(0),p)"
Next
For i:=1 To 20 Do
  AssertEquals "Truehull(i)", "Truehull(i,My.Race$)"
  AssertNull "Truehull(i,z(0))"
Next
AssertNull "Truehull(z(0))"
AssertThrows "Truehull('a')"
AssertThrows "Truehull(1,2,3)"
AssertEquals "Truehull(1,0)", "0"
AssertEquals "Truehull(0,1)", "0"


%
%  Engines
%
For i:=1 To 9
  Assert "IsNum(Engine(i).Cost.D)"
  Assert "IsNum(Engine(i).Cost.M)"
  Assert "IsNum(Engine(i).Cost.MC)"
  Assert "IsNum(Engine(i).Cost.T)"
  Assert "IsString(Engine(i).Cost.Str)"
  Assert "IsString(Engine(i).Name)"
  Assert "IsString(Engine(i).Name.Short)"
  Assert "IsNum(Engine(i).Speed$)"
  Assert "IsNum(Engine(i).Tech)"
  Assert "IsNum(Engine(i).Tech.Engine)"

  AssertEquals "Engine(i).Tech", "Engine(i).Tech.Engine"
  AssertEquals "CExtract(Engine(i).Cost.Str, 'T')", "Engine(i).Cost.T"
  AssertEquals "CExtract(Engine(i).Cost.Str, 'D')", "Engine(i).Cost.D"
  AssertEquals "CExtract(Engine(i).Cost.Str, 'M')", "Engine(i).Cost.M"
  AssertEquals "CExtract(Engine(i).Cost.Str, '$')", "Engine(i).Cost.MC"

  Assert "Engine(i).Speed$ >= 0 and Engine(i).Speed$ < 10"

  If IsV2 Then
    For w:=0 To 9 Do Assert "IsNum(Engine(i).FuelFactor(w))"
    AssertEquals "Engine(i).FuelFactor(0)", "0"
  EndIf
Next

% Test that engines are With-able and ForEach-able.
With Engine(2) Do
  AssertEquals "Tech", "Tech.Engine"
  Assert "IsNum(Tech)"
  If IsV2 Then
    AssertEquals "FuelFactor(0)", "0"
    AssertEquals "FuelFactor(9)", "Engine(2).FuelFactor(9)"
  EndIf
EndWith
ForEach Engine
  AssertEquals "Tech", "Tech.Engine"
  Assert "IsNum(Tech)"
  If IsV2 Then AssertEquals "FuelFactor(0)", "0"
Next
ForEach Global.Engine Do Assert "IsNum(Tech)"
If IsV2
  Assert "Count(Engine) = 9"
EndIf


%
%  Distance
%
Local pid, sid
pid := FindPlanet(Owner$<>0)
sid := FindShip(Owner$<>0)

% Make sure we're not bitten by FP precision problems like Tim's host is
For i:=1 To 180 Do
  AssertEquals "Distance(0,0,0,"&i&")", i
  AssertEquals "Distance(0,0,"&i&",0)", i
  AssertEquals "Int(Distance(0,0,0,"&i&"))", i
  AssertEquals "Int(Distance(0,0,"&i&",0))", i
Next

% Test distances between objects
If pid And sid Then
  AssertEquals "Distance(Planet(pid),Planet(pid))", 0
  AssertEquals "Distance(Ship(sid),Ship(sid))", 0
  AssertEquals "Distance(Ship(sid),Planet(pid))", "Distance(Planet(pid),Ship(sid))"

  AssertEquals "Distance(Planet(pid),"&Planet(pid).Loc.X&","&Planet(pid).Loc.Y&")", 0
  AssertEquals "Distance("&Planet(pid).Loc.X&","&Planet(pid).Loc.Y&",Planet(pid))", 0
Else
  Print "Not checking 'Distance(obj...)' because no objects found"
EndIf

% Test failure cases
AssertThrows "a:=Distance(1,2)"
AssertThrows "a:=Distance(1,2,3)"
AssertThrows "a:=Distance(Planet(pid),1)"
AssertThrows "a:=Distance(Planet(pid),Ship(sid),1)"
AssertThrows "a:=Distance(Planet(pid),1,Ship(sid))"
AssertThrows "a:=Distance(1,Planet(pid))"


%
%  Random
%
For i:=1 To 1000 Do
  a := Random(0,100)
  Assert "IsNum(a)"
  Assert "a < 100"
  Assert "a >= 0"
Next
For i:=1 To 1000 Do
  a := Random(100,0)
  Assert "a <= 100"
  Assert "a > 0"
Next
For i:=1 To 1000 Do
  a := Random(100)
  Assert "a < 100"
  Assert "a >= 0"
Next

% Ensure replayability
a := System.RandomSeed
b := Random(1000)
c := Random(1000)
d := Random(1000)
e := Random(1000)
f := System.RandomSeed
System.RandomSeed := a
g := Random(1000)
h := Random(1000)
i := Random(1000)
j := Random(1000)
AssertEquals "b", "g"
AssertEquals "c", "h"
AssertEquals "d", "i"
AssertEquals "e", "j"
AssertEquals "System.RandomSeed", "f"


%
%  FCodes
%
Assert "IsSpecialFCode('mkt')"
Assert "IsSpecialFCode('MKT')"
Assert "IsSpecialFCode('bums')"
Assert "not IsSpecialFCode('123')"

For i:=1 To 100 Do Assert "not IsSpecialFCode(RandomFCode())"
For i:=1 To 100 Do Assert "Len(RandomFCode()) = 3"

%
%  Minefields
%
Local Selftest.LastMinefield = 0
ForEach Minefield Do
  Assert "IsNum(Id)"
  Assert "IsNum(LastScan)"
  Assert "IsNum(Loc.X)"
  Assert "IsNum(Loc.Y)"
  Assert "IsString(Owner)"
  Assert "IsString(Owner.Adj)"
  Assert "IsNum(Owner$)"
  Assert "IsNum(Radius)"
  Assert "IsNum(Scanned)"
  Assert "IsString(Type)"
  Assert "IsNum(Units)"

  Assert "Owner$>0 and Owner$<=12"
  AssertEquals "Owner", "Player(Owner$).Race.Short"
  AssertEquals "Owner.Adj", "Player(Owner$).Race.Adj"
  Assert "Type$=0 or Type$=1"
  Assert "Units>=0"
  If IsV2 Then
    AssertEquals "Int(Sqrt(Units))", "Radius"
  Else
    AssertEquals "Round(Sqrt(Units))", "Radius"
  EndIf
  Selftest.LastMinefield := Id
Next
ForEach Global.Minefield Do Assert "IsNum(Id)"

If Selftest.LastMinefield
  % Test that minefields are With-able, indexable, and findable.
  With Minefield(Selftest.LastMinefield)
    AssertEquals "Id", "Selftest.LastMinefield"
    Assert "Type$=0 or Type$=1"
  EndWith
  AssertEquals "Minefield(Selftest.LastMinefield).Id", "Selftest.LastMinefield"
  Assert "Minefield(Selftest.LastMinefield).Type$=0 or Minefield(Selftest.LastMinefield).Type$=1"
  If IsV2 Then
    AssertEquals "Find(Minefield, Id=" & Selftest.LastMinefield & ", Id)", Selftest.LastMinefield
    AssertEquals "Count(Minefield, Id=" & Selftest.LastMinefield & ")", 1
    Assert "Count(Minefield) > 0"
  EndIf
Else
  Print "Not checking minefields, none present"
  If IsV2 Then AssertEquals "Count(Minefield)", "0"
EndIf


%
%  Ion Storms
%
Local Selftest.LastStorm = 0
ForEach Storm Do
  Assert "IsNum(Class)"
  Assert "IsString(Heading)"
  Assert "IsNum(Heading$)"
  Assert "IsNum(Id)"
  Assert "IsNum(Loc.X)"
  Assert "IsNum(Loc.Y)"
  Assert "IsString(Name)"
  Assert "IsNum(Radius)"
  Assert "IsString(Speed)"
  Assert "IsNum(Speed$)"
  Assert "IsString(Status)"
  Assert "Status$=0 or Status$=1"
  Assert "IsNum(Voltage)"

  Assert "Class>0 and Class<=5"
  Assert "Heading$>=0 and Heading$<360"
  Assert "Radius>0"
  Assert "Speed$>0"
  Assert "InStr(Speed, Speed$&'')"
  Assert "Voltage>0"

  Select Case Class
    Case 1
      Assert "Voltage<50"
    Case 2
      Assert "Voltage>=50 and Voltage<100"
    Case 3
      Assert "Voltage>=100 and Voltage<150"
    Case 4
      Assert "Voltage>=150 and Voltage<200"
    Case 5
      Assert "Voltage>=200"
  EndSelect

  Selftest.LastStorm := Id
Next
ForEach Global.Storm Do Assert "IsNum(Id)"

If Selftest.LastStorm Then
  % Test that storms are With-able, indexable, and findable.
  With Storm(Selftest.LastStorm)
    AssertEquals "Id", "Selftest.LastStorm"
    Assert "Class>0 and Class<=5"
  EndWith
  AssertEquals "Storm(Selftest.LastStorm).Id", "Selftest.LastStorm"
  Assert "Storm(Selftest.LastStorm).Class>0 and Storm(Selftest.LastStorm).Class<=5"
  If IsV2 Then
    AssertEquals "Find(Storm, Id=" & Selftest.LastStorm & ", Id)", Selftest.LastStorm
    AssertEquals "Count(Storm, Id=" & Selftest.LastStorm & ")", 1
    Assert "Count(Storm) > 0"
  EndIf
Else
  Print "Not checking ion storms, none present"
  If IsV2 Then AssertEquals "Count(Storm)", "0"
EndIf


%
%  Ufos
%
Local Selftest.LastUfo = 0
ForEach Ufo Do
  Assert "IsNum(Color)"
  Assert "IsNum(Color.EGA)"
  Assert "IsString(Heading) or IsEmpty(Heading)"
  Assert "IsNum(Heading$) or IsEmpty(Heading$)"
  Assert "IsNum(Id)"
  Assert "IsNum(Id2)"
  Assert "IsString(Info1)"
  Assert "IsString(Info2)"
  Assert "Keep=0 or Keep=1"
  Assert "IsNum(LastScan)"
  Assert "IsNum(Loc.X)"
  Assert "IsNum(Loc.Y)"
  Assert "IsNum(Move.DX)"
  Assert "IsNum(Move.DY)"
  Assert "IsString(Name)"
  Assert "IsNum(Radius)"
  Assert "IsString(Speed)"
  Assert "IsNum(Speed$)"
  Assert "IsNum(Type)"
  Assert "IsNum(Visible.Planet)"
  Assert "IsNum(Visible.Ship)"

  AssertEquals "IsEmpty(Heading)", "IsEmpty(Heading$)"
  Assert "InStr(Speed, Speed$&'')"

  If Not Selftest.LastUfo Then Selftest.LastUfo := Id
Next
ForEach Global.Ufo Do Assert "IsNum(Id)"
If Not Selftest.LastUfo
  Print "Not checking Ufos, none found"
  If IsV2 Then Assert "Count(Ufo)=0"
Else
  % Test that Ufos are With-able, indexable, and findable.
  With Ufo(Selftest.LastUfo)
    AssertEquals "Id", "Selftest.LastUfo"
  EndWith
  AssertEquals "Ufo(Selftest.LastUfo).Id", "Selftest.LastUfo"
  If IsV2 Then
    AssertEquals "Find(Ufo, Id=" & Selftest.LastUfo & ", Id)", Selftest.LastUfo
    Assert "Count(Ufo, Id=" & Selftest.LastUfo & ") >= 1"
    Assert "Count(Ufo) > 0"
  EndIf
EndIf


%
%  Find/Count
%
AssertNull "FindShip(0)"
AssertNull "FindPlanet(0)"
AssertEquals "CountShips(0)", "0"
AssertEquals "CountPlanets(0)", "0"

If IsV2 Then
  AssertNull "Find(Ship,0,Id)"
  AssertNull "Find(Planet,0,Id)"
  AssertEquals "Count(Ship,0)", "0"
  AssertEquals "Count(Planet,0)", "0"
EndIf

% The 'FoldJump' optimisation originally failed on these:
If FindShip(Owner$<>0) Then
  Assert "1"
EndIf
If Not FindShip(Owner$<>0) Then
  Assert "1"
EndIf
If IsEmpty(FindShip(Owner$<>0))
  Assert "1"
EndIf
If Not IsEmpty(FindShip(Owner$<>0))
  Assert "1"
EndIf
a:=If(FindShip(Owner$<>0),17,4)
Assert "a"

%
%  Ship
%
Local Selftest.DidOwnShips = 0
Local Selftest.DidOwnCarriers = 0
Local Selftest.DidOwnTorpers = 0
Local Selftest.DidOrbits = 0
Local Selftest.DidFreeSpace = 0
Local Selftest.ThisIsFleet
ForEach Ship Do
  % Check whether this is a fleet member
  Selftest.ThisIsFleet := 0
  Try Selftest.ThisIsFleet := Fleet$

  % Aux
  Assert "IsEmpty(Aux) or IsString(Aux)"
  Assert "IsEmpty(Aux$) or IsNum(Aux$)"
  Assert "IsEmpty(Aux.Ammo) or IsNum(Aux.Ammo)"
  Assert "IsEmpty(Aux.Count) or IsNum(Aux.Count)"
  Assert "IsEmpty(Aux.Short) or IsString(Aux.Short)"
  If Owner=My.Race And Aux.Count Then Assert "Aux$ >= 0"
  If Not IsEmpty(Aux$) And Aux$>0 And Aux$<=10 Then AssertEquals "Aux", "Torpedo(Aux$).Name"

  % Beam
  Assert "IsEmpty(Beam) or IsString(Beam)"
  Assert "IsEmpty(Beam$) or IsNum(Beam$)"
  Assert "IsEmpty(Beam.Count) or IsNum(Beam.Count)"
  Assert "IsEmpty(Beam.Max) or IsNum(Beam.Max)"
  Assert "IsEmpty(Beam.Short) or IsString(Beam.Short)"
  If Owner$=My.Race$ Then
    Assert "not IsEmpty(Beam.Count)"
    If Beam$ Then AssertEquals "Beam", GlobalPrefix & "Beam(Beam$).Name"
    Selftest.DidOwnShips := 1
  EndIf

  % Cargo
  If Owner$=My.Race$ Then
    Assert "IsNum(Cargo.Colonists)"
    Assert "IsNum(Cargo.T)"
    Assert "IsNum(Cargo.D)"
    Assert "IsNum(Cargo.M)"
    Assert "IsNum(Cargo.Supplies)"
    Assert "IsNum(Cargo.Free)"
    Assert "IsNum(Cargo.Max)"
    AssertEquals "Cargo.T + Cargo.D + Cargo.M + Cargo.Colonists + Cargo.Supplies + If(Aux$, Aux.Ammo, 0) + Cargo.Free", "Cargo.Max"
    AssertEquals "Cargo.Max", GlobalPrefix & "Hull(Hull$).Cargo.Max"

    Assert "IsNum(Cargo.N)"
    Assert "IsNum(Cargo.MaxFuel)"
    AssertEquals "Cargo.MaxFuel", GlobalPrefix & "Hull(Hull$).Cargo.MaxFuel"

    Assert "IsNum(Cargo.Money)"

    AssertEquals "CExtract(Cargo.Str, 'T')", "Cargo.T"
    AssertEquals "CExtract(Cargo.Str, 'D')", "Cargo.D"
    AssertEquals "CExtract(Cargo.Str, 'M')", "Cargo.M"
    AssertEquals "CExtract(Cargo.Str, 'N')", "Cargo.N"
    AssertEquals "CExtract(Cargo.Str, 'S')", "Cargo.Supplies"
    AssertEquals "CExtract(Cargo.Str, '$')", "Cargo.Money"
  EndIf

  % Comment
  c := Comment
  Comment := 9
  AssertEquals "Comment", "9"

  % SetComment stringifies. We don't check for exact results, because
  % there's comment.q which will mangle comments.
  SetComment 12
  Assert "IsString(Comment)"
  AssertEquals "Left(Comment,2)", "'12'"

  % SetComment keeps empty strings. It used to make them empty, and
  % comment.q still does.
  SetComment ""
  Assert "IsEmpty(Comment) or (IsString(Comment) and Comment='')"

  Comment := c

  % Crew
  Assert "IsNum(Crew) or IsEmpty(Crew)"
  If Hull Then
    Assert "IsNum(Crew.Normal)"
    AssertEquals "Crew.Normal", GlobalPrefix & "Hull(Hull$).Crew.Normal"
  Else
    Assert "Owner$<>My.Race$"
  EndIf

  % Damage
  Assert "IsNum(Damage) or IsEmpty(damage)"
  If Owner$=My.Race$ Then Assert "Damage >= 0"

  % Enemy
  If Owner$=My.Race$ Then
    Assert "IsNum(Enemy$)"
    If Enemy$ Then
      AssertEquals "Enemy",     "Player(Enemy$).Race.Short"
      AssertEquals "Enemy.Adj", "Player(Enemy$).Race.Adj"
    Else
      AssertNull "Enemy"
      AssertNull "Enemy.Adj"
    EndIf

    e := Enemy$
    SetEnemy 3
    AssertEquals "Enemy$", 3
    AssertEquals "Enemy", "Player(3).Race.Short"
    SetEnemy 0
    AssertNull "Enemy"
    AssertNull "Enemy.Adj"
    Enemy$ := 5
    AssertEquals "Enemy$", 5
    AssertEquals "Enemy", "Player(5).Race.Short"
    SetEnemy e
    AssertEquals "e", "Enemy$"
  EndIf

  % Engine
  If Owner$=My.Race$ Or Engine Then
    Assert "IsNum(Engine$)"
    AssertEquals "Engine", GlobalPrefix & "Engine(Engine$).Name"
  EndIf
  If Owner$=My.Race$ Or Hull Then
    Assert "IsNum(Engine.Count)"
    AssertEquals "Engine.Count", GlobalPrefix & "Hull(Hull$).Engine.Count"
  EndIf

  % FCode
  Assert "IsString(FCode) or IsEmpty(FCode)"
  If Owner$=My.Race$ Then
    f := FCode
    SetFCode "abc"
    AssertEquals "FCode", "'abc'"
    SetFCode Z(0)
    AssertEquals "FCode", "'abc'"
    If IsV2 Then FCode := Z(0)             % FIXME: this is an error in 1.x. Is this better?
    AssertEquals "FCode", "'abc'"
    SetFCode "toolong"
    AssertEquals "FCode", "'too'"
    FCode := f
    AssertEquals "FCode", "f"
  EndIf

  % Fighter
  Assert "IsNum(Fighter.Bays) or IsEmpty(Fighter.Bays)"
  Assert "IsNum(Fighter.Count) or IsEmpty(Fighter.Count)"
  If Owner$=My.Race$ Then
    If Fighter.Bays Then
      AssertEquals "Aux$", "11"
      AssertEquals "Aux.Ammo", "Fighter.Count"
      AssertEquals "Aux.Count", "Fighter.Bays"
      Selftest.DidOwnCarriers := 1
    EndIf
    AssertEquals "Fighter.Bays", GlobalPrefix & "Hull(Hull$).Fighter.Bays"
  EndIf

  % Heading
  Assert "IsString(Heading) or IsEmpty(Heading)"
  Assert "IsNum(Heading$) or IsEmpty(Heading$)"
  If Owner$=My.Race$ Then
    If Waypoint.DX Or Waypoint.DY Then
      Assert "Heading"
      Assert "Heading$ >= 0"
    Else
      AssertNull "Heading"
      AssertNull "Heading$"
    EndIf
  EndIf

  % Hull
  Assert "IsString(Hull) or IsEmpty(Hull)"
  Assert "IsNum(Hull$) or IsEmpty(Hull$)"
  Assert "IsString(Hull.Short) or IsEmpty(Hull.Short)"
  Assert "IsString(Hull.Special) or IsEmpty(Hull.Special)"
  If Owner$=My.Race$ Or Hull Then
    AssertEquals "Hull", GlobalPrefix & "Hull(Hull$).Name"
    AssertEquals "Hull.Short", GlobalPrefix & "Hull(Hull$).Name.Short"
    % This may differ thanks to ship-specific abilities; Hull.Special is null for foreign ships in v1
    % AssertEquals "Hull.Special", GlobalPrefix & "Hull(Hull$).Special"
    AssertEquals "Tech.Hull", GlobalPrefix & "Hull(Hull$).Tech"
  EndIf

  % Id
  Assert "IsNum(Id)"
  Assert "Id > 0 and Id < 1000"

  % Level
  Assert "IsEmpty(Level) or IsNum(Level)"
  If Cfg("NumExperienceLevels") And Owner$=My.Race$ Then Assert "IsNum(Level) and Level >= 0"

  % Loc
  Assert "IsString(Loc)"
  Assert "IsNum(Loc.X)"
  Assert "IsNum(Loc.Y)"

  % Mark/Unmark
  M := Marked
  Assert "m=0 or m=1"
  Mark
  Assert "Marked"
  Unmark
  Assert "not Marked"
  % 'Mark Z(0)' means 'Mark 1' by the rule that an omitted argument is the same as a null argument.
  Mark Z(0)
  Assert "Marked"
  Mark 0
  Assert "not Marked"
  Mark 1
  Assert "Marked"
  Mark M
  AssertEquals "m", "Marked"

  % Mass
  Assert "IsEmpty(Mass) or IsNum(Mass)"
  If Owner$=My.Race$ Then
    Assert "IsNum(Mass)"
    % FIXME: cross-check that mass is computed right?
  EndIf

  % Mission
  Assert "IsEmpty(Mission) or IsString(Mission)"
  Assert "IsEmpty(Mission$) or IsNum(Mission$)"
  Assert "IsEmpty(Mission.Intercept) or IsNum(Mission.Intercept)"
  Assert "IsEmpty(Mission.Short) or IsString(Mission.Short)"
  Assert "IsEmpty(Mission.Tow) or IsNum(Mission.Tow)"
  If Owner$=My.Race$ Then
    Assert "IsString(Mission)"
    Assert "IsNum(Mission$)"
    Assert "IsNum(Mission.Intercept)"
    Assert "IsNum(Mission.Tow)"
    Assert "IsString(Mission.Short)"

    M := Mission$
    i := Mission.Intercept
    T := Mission.Tow

    SetMission 4
    AssertEquals "Mission$", "4"
    AssertEquals "Mission.Intercept", "0"
    AssertEquals "Mission.Tow", "0"

    SetMission 7,0,Id
    AssertEquals "Mission$", "7"
    AssertEquals "Mission.Intercept", "0"
    AssertEquals "Mission.Tow", "Id"

    SetMission 777,888,999
    AssertEquals "Mission$", "777"
    AssertEquals "Mission.Intercept", "888"
    AssertEquals "Mission.Tow", "999"

    SetMission Z(0),9,8
    AssertEquals "Mission$", "777"
    AssertEquals "Mission.Intercept", "888"
    AssertEquals "Mission.Tow", "999"

    SetMission 222,Z(0),3
    AssertEquals "Mission$", "222"
    AssertEquals "Mission.Intercept", "0"
    AssertEquals "Mission.Tow", "3"

    SetMission M,i,T
    AssertEquals "Mission$", "m"
    AssertEquals "Mission.Intercept", "i"
    AssertEquals "Mission.Tow", "t"
  EndIf

  % Name
  If Not IsEmpty(Name)
    Assert "IsString(Name)"
    N := Name
    Name := "hi there"
    AssertEquals "Name", "'hi there'"

    % SetName on foreign ships is permitted since 1.1.18
    SetName "hoho"
    AssertEquals "Name", "'hoho'"
    SetName Z(0)
    AssertEquals "Name", "'hoho'"
    SetName N

    AssertEquals "Name", "n"
  Else
    Assert "Owner$<>My.Race$"
  EndIf

  % Orbit
  If Orbit
    Assert "IsString(Orbit)"
    Assert "IsNum(Orbit$)"
    Assert "Orbit$ > 0 and Orbit$ <= 500"
    AssertEquals "Orbit", "Planet(Orbit$).Name"
    AssertEquals "Loc.X", "Planet(Orbit$).Loc.X"
    AssertEquals "Loc.Y", "Planet(Orbit$).Loc.Y"
    AssertEquals "PlanetAt(Loc.X, Loc.Y)", "Orbit$"
    Selftest.DidOrbits := 1
  Else
    AssertEquals "Orbit$", "0"
    AssertNull "FindPlanet(Loc.X = Ship.Loc.X and Loc.Y = Ship.Loc.Y)"
    AssertEquals "PlanetAt(Loc.X, Loc.Y)", "0"
    Selftest.DidFreeSpace := 1
  EndIf

  % Owner
  Assert "IsString(Owner)"
  Assert "IsNum(Owner$)"
  Assert "IsString(Owner.Adj)"
  Assert "IsNum(Owner.Real)"
  Assert "Owner$>0 and Owner$<=12"
  AssertEquals "Owner", "Player(Owner$).Race.Short"
  AssertEquals "Owner.Adj", "Player(Owner$).Race.Adj"

  % Speed
  Assert "IsString(Speed) or IsEmpty(Speed)"
  Assert "IsNum(Speed$) or IsEmpty(Speed$)"
  If Owner$=My.Race$ Then
    Assert "IsNum(Speed$)"
    If Not Selftest.ThisIsFleet
      s := Speed$
      SetSpeed 3
      AssertEquals "Speed$", "3"
      SetSpeed Z(0)
      AssertEquals "Speed$", "3"
      Assert "InStr(Speed, '3')"
      Speed$ := 9
      AssertEquals "Speed$", "9"
      Assert "InStr(Speed, '9')"
      SetSpeed 0
      AssertEquals "Speed$", "0"
      SetSpeed s
      AssertEquals "Speed$", "s"
    EndIf
  EndIf

  % Torp
  Assert "IsEmpty(Torp) or IsString(Torp)"
  Assert "IsEmpty(Torp$) or IsNum(Torp$)"
  Assert "IsEmpty(Torp.Count) or IsNum(Torp.Count)"
  Assert "IsEmpty(Torp.LCount) or IsNum(Torp.LCount)"
  Assert "IsEmpty(Torp.LMax) or IsNum(Torp.LMax)"
  Assert "IsEmpty(Torp.Short) or IsString(Torp.Short)"
  If Owner$=My.Race$ And Torp Then
    Assert "Torp$<>0"
    Assert "Torp.LCount<>0"
    Assert "Torp.LMax<>0"
    Assert "Torp.LCount <= Torp.LMax"
    AssertEquals "Torp", "Torpedo(Torp$).Name"
    AssertEquals "Torp.Short", "Torpedo(Torp$).Name.Short"
    AssertEquals "Torp.LMax", GlobalPrefix & "Hull(Hull$).Torp.LMax"
    AssertEquals "Torp$", "Aux$"
    AssertEquals "Torp", "Aux"
    AssertEquals "Torp.LCount", "Aux.Count"
    AssertEquals "Torp.Count", "Aux.Ammo"
    Selftest.DidOwnTorpers := 1
  EndIf

  % Transfer.Ship - simple type checks only
  Assert "IsEmpty(Transfer.Ship) or Transfer.Ship=0 or Transfer.Ship=1"
  Assert "IsEmpty(Transfer.Ship.Colonists) or IsNum(Transfer.Ship.Colonists)"
  Assert "IsEmpty(Transfer.Ship.D) or IsNum(Transfer.Ship.D)"
  Assert "IsEmpty(Transfer.Ship.Id) or IsNum(Transfer.Ship.Id)"
  Assert "IsEmpty(Transfer.Ship.M) or IsNum(Transfer.Ship.M)"
  Assert "IsEmpty(Transfer.Ship.N) or IsNum(Transfer.Ship.N)"
  Assert "IsEmpty(Transfer.Ship.Name) or IsString(Transfer.Ship.Name)"
  Assert "IsEmpty(Transfer.Ship.Supplies) or IsNum(Transfer.Ship.Supplies)"
  Assert "IsEmpty(Transfer.Ship.T) or IsNum(Transfer.Ship.T)"

  % Transfer.Unload - simple type checks only
  Assert "IsEmpty(Transfer.Unload) or Transfer.Unload=0 or Transfer.Unload=1"
  Assert "IsEmpty(Transfer.Unload.Colonists) or IsNum(Transfer.Unload.Colonists)"
  Assert "IsEmpty(Transfer.Unload.D) or IsNum(Transfer.Unload.D)"
  Assert "IsEmpty(Transfer.Unload.Id) or IsNum(Transfer.Unload.Id)"
  Assert "IsEmpty(Transfer.Unload.M) or IsNum(Transfer.Unload.M)"
  Assert "IsEmpty(Transfer.Unload.N) or IsNum(Transfer.Unload.N)"
  Assert "IsEmpty(Transfer.Unload.Name) or IsString(Transfer.Unload.Name)"
  Assert "IsEmpty(Transfer.Unload.Supplies) or IsNum(Transfer.Unload.Supplies)"
  Assert "IsEmpty(Transfer.Unload.T) or IsNum(Transfer.Unload.T)"

  % Type
  If Type
    Assert "IsString(Type)"
    Assert "IsString(Type.Short)"
    AssertEquals "Mid(Type,1,1)", "Type.Short"
    Assert "Hull"
  Else
    AssertNull "Type"
    AssertNull "Type.Short"
  EndIf

  % Waypoint, Move
  If Owner$=My.Race$ Then
    Assert "IsString(Waypoint)"
    Assert "IsNum(Waypoint.Dist)"
    Assert "IsNum(Waypoint.DX)"
    Assert "IsNum(Waypoint.DY)"
    Assert "IsNum(Waypoint.Planet)"
    Assert "IsNum(Waypoint.X)"
    Assert "IsNum(Waypoint.Y)"
    If Speed$ Then
      Assert "IsNum(Move.Eta)"
      Assert "IsNum(Move.Fuel)"
    EndIf
    If Waypoint.Planet Then
      % Cannot check these, as there might be wrap. Cfg("AllowWraparoundMap") works for PHost, not THost.
      % AssertEquals "Waypoint.X", "Planet(Waypoint.Planet).Loc.X"
      % AssertEquals "Waypoint.Y", "Planet(Waypoint.Planet).Loc.Y"
    Else
      AssertNull "FindPlanet(Loc.X = Waypoint.X and Loc.Y = Waypoint.Y)"
    EndIf

    % Without the 'Int', this comparison fails in v1 and v2. Floats are compared exactly,
    % there is no 1.0e-6 "fuzz" allowed like for conversion to bool. However, without our
    % own fuzz factor here, this test fails for DX=36, DY=-17
    AssertEquals "Waypoint.DX^2 + Waypoint.DY^2", "int(Waypoint.Dist^2 + 0.0000001)"
    Assert "not (Waypoint.DX^2 + Waypoint.DY^2 - Waypoint.Dist^2)"

    If Mission$<>8 And Not Selftest.ThisIsFleet Then
      % don't check this when we're intercepting
      X := Waypoint.X
      Y := Waypoint.Y

      SetWaypoint Loc.X, Loc.Y
      AssertEquals "Waypoint.Planet", "Orbit$"
      AssertEquals "Waypoint.X", "Loc.X"
      AssertEquals "Waypoint.Y", "Loc.Y"
      AssertEquals "Waypoint.DX", "0"
      AssertEquals "Waypoint.DY", "0"
      AssertEquals "Waypoint.Dist", "0"
      If Speed$ Then
        AssertEquals "Move.Eta", "0"
        AssertEquals "Move.Fuel", "0"
      EndIf

      SetWaypoint Loc.X, Loc.Y + 1
      If Orbit$ Then Assert "Waypoint.Planet <> Orbit$"
      AssertEquals "Waypoint.X", "Loc.X"
      AssertEquals "Waypoint.Y", "Loc.Y+1"
      AssertEquals "Waypoint.DX", "0"
      AssertEquals "Waypoint.DY", "1"
      AssertEquals "Waypoint.Dist", "1"
      If Speed$ Then AssertEquals "Move.Eta", "1"

      SetWaypoint X, Y
      AssertEquals "Waypoint.X", "X"
      AssertEquals "Waypoint.Y", "Y"
    EndIf
  EndIf
Next
ForEach Global.Ship Do Assert "IsNum(Id)"

% Test that ships are With-able, indexable, and findable.
sid := FindShip(Owner$<>0)
With Ship(sid) Do
  AssertEquals "Id", "sid"
  Assert "Owner$<>0"
EndWith
AssertEquals "Ship(sid).Id", "sid"
Assert "Ship(sid).Owner$<>0"

If IsV2 Then
  Assert "Count(Ship) > 0"
  AssertEquals "Find(Ship, Owner$<>0, Id)", "sid"
EndIf

% Warn about things we didn't check
If Not Selftest.DidOwnShips    Then Print "Not checking various ship properties, no own ships found"
If Not Selftest.DidOwnCarriers Then Print "Not checking various carrier properties, no own carriers found"
If Not Selftest.DidOwnTorpers  Then Print "Not checking various torper properties, no own torpers found"
If Not Selftest.DidOrbits      Then Print "Not checking orbit relations, no ships in orbit found"
If Not Selftest.DidFreeSpace   Then Print "Not checking free space ships, no ships in free space found"


%
%  Planets
%
Local Selftest.DidOwnPlanet = 0
Local Selftest.DidBuildBase = 0
Local Selftest.DidBuildStructures = 0
Local Selftest.DidBase = 0
Local Selftest.DidShipyard = 0
ForEach Planet Do
  % Base status
  If Owner$=My.Race$
    Selftest.DidOwnPlanet := 1
    Assert "IsString(Base)"
    Assert "Base.YesNo=0 or Base.YesNo=1"

    T:=Mined.T; D:=Mined.D; M:=Mined.M; C:=Money
    If Base = "being built"
      % Try to cancel the base, and rebuild it
      BuildBase 0
      Assert "Mined.T > T"
      Assert "Mined.D > D"
      Assert "Mined.M > M"
      Assert "Money > C"
      BuildBase 1
      AssertEquals "Mined.T", "T"
      AssertEquals "Mined.D", "D"
      AssertEquals "Mined.M", "M"
      AssertEquals "Money", "C"
      Selftest.DidBuildBase := 1
    Else If Base.YesNo
      % There is a base, check that BuildBase fails
      AssertThrows "BuildBase"
    Else
      % No base, try building one
      Try
        BuildBase 1
        Assert "Mined.T < T"
        Assert "Mined.D < D"
        Assert "Mined.M < M"
        Assert "Money <= C"
        BuildBase 0
        AssertEquals "Mined.T", "T"
        AssertEquals "Mined.D", "D"
        AssertEquals "Mined.M", "M"
        If Money > c Then SellSupplies c - Money
        AssertEquals "Money", "c"
        Selftest.DidBuildBase := 1
      EndTry
    EndIf
  EndIf

  % Structures
  Assert "IsNum(Defense.Want)"
  Assert "IsNum(Defense.Base.Want)"
  Assert "IsNum(Factories.Want)"
  Assert "IsNum(Mines.Want)"
  If Owner$=My.Race$ Then
    Assert "IsNum(Defense)"
    Assert "IsNum(Factories)"
    Assert "IsNum(Mines)"
    If Base.YesNo
      Assert "IsNum(Defense.Base)"
      Assert "IsNum(Defense.Base.Max)"
    Else
      AssertNull "Defense.Base"
      AssertNull "Defense.Base.Max"
    EndIf
    Assert "IsNum(Defense.Max)"
    Assert "IsNum(Factories.Max)"
    Assert "IsNum(Mines.Max)"

    % Try some building. This does intentionally refuse building using supply sale,
    % so we have one less thing to undo.
    % FIXME: this does not test the 'n' forms of the statements
    f := Factories
    M := Mines
    D := Defense
    b := Defense.Base
    If Supplies>0 And Money>=3 And Factories<Factories.Max Then
      % Factories
      BuildFactories 1
      AssertEquals "Factories", "f+1"
      BuildFactories 0
      AssertEquals "Factories", "f+1"
      BuildFactories Z(0)
      AssertEquals "Factories", "f+1"
      BuildFactories -1
      AssertEquals "Factories", "f"
      Selftest.DidBuildStructures := BitOr(Selftest.DidBuildStructures, 1)
    EndIf

    If Supplies>0 And Money>=4 And Mines<Mines.Max Then
      % Mines
      BuildMines 1
      AssertEquals "Mines", "m+1"
      BuildMines 0
      AssertEquals "Mines", "m+1"
      BuildMines Z(0)
      AssertEquals "Mines", "m+1"
      BuildMines -1
      AssertEquals "Mines", "m"
      Selftest.DidBuildStructures := BitOr(Selftest.DidBuildStructures, 2)
    EndIf

    If Supplies>0 And Money>=10 And Defense<Defense.Max Then
      % Defense
      BuildDefense 1
      AssertEquals "Defense", "d+1"
      BuildDefense 0
      AssertEquals "Defense", "d+1"
      BuildDefense Z(0)
      AssertEquals "Defense", "d+1"
      BuildDefense -1
      AssertEquals "Defense", "d"
      Selftest.DidBuildStructures := BitOr(Selftest.DidBuildStructures, 4)
    EndIf

    If Mined.D>0 And Money>=10 And Base.YesNo And Defense.Base<Defense.Base.Max Then
      % Base Defense
      BuildBaseDefense 1
      AssertEquals "Defense.Base", "b+1"
      BuildBaseDefense 0
      AssertEquals "Defense.Base", "b+1"
      BuildBaseDefense Z(0)
      AssertEquals "Defense.Base", "b+1"
      BuildBaseDefense -1
      AssertEquals "Defense.Base", "b"
      Selftest.DidBuildStructures := BitOr(Selftest.DidBuildStructures, 8)
    EndIf
  Else
    Assert "IsEmpty(Defense) or IsNum(Defense)"
    Assert "IsEmpty(Factories) or IsNum(Factories)"
    Assert "IsEmpty(Mines) or IsNum(Mines)"
    Assert "IsEmpty(Defense.Base) or IsNum(Defense.Base)"
    Assert "IsEmpty(Defense.Max) or IsNum(Defense.Max)"
    Assert "IsEmpty(Defense.Base.Max) or IsNum(Defense.Base.Max)"
    Assert "IsEmpty(Factories.Max) or IsNum(Factories.Max)"
    Assert "IsEmpty(Mines.Max) or IsNum(Mines.Max)"
    AssertThrows "BuildBaseDefense 1"
    AssertThrows "BuildDefense 1"
    AssertThrows "BuildFactories 1"
    AssertThrows "BuildMines 1"
    AssertThrows "AutoBuild"
  EndIf

  % Colonists
  If Owner$=My.Race$ Then
    Assert "IsNum(Colonists)"
    Assert "IsString(Colonists.Change)"
    Assert "IsNum(Colonists.Change$)"
    Assert "IsString(Colonists.Happy)"
    Assert "IsNum(Colonists.Happy$)"
    Assert "IsNum(Colonists.Supported)"
    Assert "IsNum(Colonists.Tax)"

    T := Colonists.Tax
    SetColonistTax 3
    AssertEquals "Colonists.Tax", "3"
    h := Colonists.Change$
    SetColonistTax Z(0)
    AssertEquals "Colonists.Tax", "3"
    Colonists.Tax := 5
    AssertEquals "Colonists.Tax", "5"
    Colonists.Tax := 0
    AssertEquals "Colonists.Tax", "0"
    Colonists.Tax := 50
    AssertEquals "Colonists.Tax", "50"
    Assert "Colonists.Change$ < h"
    AssertThrows "SetColonistTax 101"
    AssertThrows "SetColonistTax -1"
    AssertThrows "Colonist.Tax := 101"
    SetColonistTax T
    AssertEquals "Colonists.Tax", "T"
  Else
    Assert "IsEmpty(Colonists) or IsNum(Colonists)"
    Assert "IsEmpty(Colonists.Change) or IsString(Colonists.Change)"
    Assert "IsEmpty(Colonists.Change$) or IsNum(Colonists.Change$)"
    Assert "IsEmpty(Colonists.Happy) or IsString(Colonists.Happy)"
    Assert "IsEmpty(Colonists.Happy$) or IsNum(Colonists.Happy$)"
    Assert "IsEmpty(Colonists.Supported) or IsNum(Colonists.Supported)"
    Assert "IsEmpty(Colonists.Tax) or IsNum(Colonists.Tax)"
    AssertThrows "SetColonistTax 1"
  EndIf

  % Comment (copied from ship)
  c := Comment
  Comment := 9
  AssertEquals "Comment", "9"

  % SetComment stringifies. We don't check for exact results, because
  % there's comment.q which will mangle comments.
  SetComment 12
  Assert "IsString(Comment)"
  AssertEquals "Left(Comment,2)", "'12'"

  % SetComment keeps empty strings. It used to make them empty, and
  % comment.q still does.
  SetComment ""
  Assert "IsEmpty(Comment) or (IsString(Comment) and Comment='')"

  Comment := c

  % Density
  If Owner$=My.Race$ Then
    Assert "IsNum(Density.N)"
    Assert "IsNum(Density.T)"
    Assert "IsNum(Density.D)"
    Assert "IsNum(Density.M)"
  Else
    Assert "IsNum(Density.N) or IsEmpty(Density.N)"
    Assert "IsNum(Density.T) or IsEmpty(Density.T)"
    Assert "IsNum(Density.D) or IsEmpty(Density.D)"
    Assert "IsNum(Density.M) or IsEmpty(Density.M)"
  EndIf

  % FCode (copied from ship)
  Assert "IsString(FCode) or IsEmpty(FCode)"
  If Owner$=My.Race$ Then
    f := FCode
    SetFCode "abc"
    AssertEquals "FCode", "'abc'"
    SetFCode Z(0)
    AssertEquals "FCode", "'abc'"
    If IsV2 Then FCode := Z(0)             % FIXME: this is an error in 1.x. Is this better?
    AssertEquals "FCode", "'abc'"
    SetFCode "toolong"
    AssertEquals "FCode", "'too'"
    FCode := f
    AssertEquals "FCode", "f"
  EndIf

  % Ground
  If Owner$=My.Race$ Then
    Assert "IsNum(Ground.N)"
    Assert "IsNum(Ground.T)"
    Assert "IsNum(Ground.D)"
    Assert "IsNum(Ground.M)"
  Else
    Assert "IsNum(Ground.N) or IsEmpty(Ground.N)"
    Assert "IsNum(Ground.T) or IsEmpty(Ground.T)"
    Assert "IsNum(Ground.D) or IsEmpty(Ground.D)"
    Assert "IsNum(Ground.M) or IsEmpty(Ground.M)"
  EndIf

  % Id (copied from ship)
  Assert "IsNum(Id)"
  Assert "Id > 0 and Id <= 500"

  % Industry
  If Industry Or Owner$=My.Race$
    Assert "IsString(Industry)"
    Assert "IsNum(Industry$)"
  Else
    AssertNull "Industry"
    AssertNull "Industry$"
  EndIf

  % Level (copied from ship)
  Assert "IsEmpty(Level) or IsNum(Level)"
  If Cfg("NumExperienceLevels") And Owner$=My.Race$ Then Assert "IsNum(Level) and Level >= 0"

  % Loc (copied from ship)
  Assert "IsNum(Loc.X)"
  Assert "IsNum(Loc.Y)"
  AssertEquals "PlanetAt(Loc.X, Loc.Y)", "Id"

  % Mark/Unmark (copied from ship)
  M := Marked
  Assert "m=0 or m=1"
  Mark
  Assert "Marked"
  Unmark
  Assert "not Marked"
  % 'Mark Z(0)' means 'Mark 1' by the rule that an omitted argument is the same as a null argument.
  Mark Z(0)
  Assert "Marked"
  Mark 0
  Assert "not Marked"
  Mark 1
  Assert "Marked"
  Mark M
  AssertEquals "m", "Marked"

  % Mined
  If Owner$=My.Race$ Then
    Assert "IsNum(Mined.N)"
    Assert "IsNum(Mined.T)"
    Assert "IsNum(Mined.D)"
    Assert "IsNum(Mined.M)"
    Assert "IsString(Mined.Str)"
    AssertEquals "CExtract(Mined.Str, 'N')", "Mined.N"
    AssertEquals "CExtract(Mined.Str, 'T')", "Mined.T"
    AssertEquals "CExtract(Mined.Str, 'D')", "Mined.D"
    AssertEquals "CExtract(Mined.Str, 'M')", "Mined.M"
  Else
    Assert "IsNum(Mined.N) or IsEmpty(Mined.N)"
    Assert "IsNum(Mined.T) or IsEmpty(Mined.T)"
    Assert "IsNum(Mined.D) or IsEmpty(Mined.D)"
    Assert "IsNum(Mined.M) or IsEmpty(Mined.M)"
  EndIf

  % Name
  Assert "IsString(Name)"

  % Natives
  If Natives And Owner$=My.Race$ Then
    % when this is our planet, everything must be known
    Assert "IsNum(Natives)"
    Assert "IsString(Natives.Change)"
    Assert "IsNum(Natives.Change$)"
    Assert "IsString(Natives.Gov)"
    Assert "IsNum(Natives.Gov$)"
    Assert "IsString(Natives.Happy)"
    Assert "IsNum(Natives.Happy$)"
    Assert "IsString(Natives.Race)"
    Assert "IsNum(Natives.Race$)"
    Assert "IsNum(Natives.Tax)"

    T := Natives.Tax
    SetNativeTax 3
    AssertEquals "Natives.Tax", "3"
    h := Natives.Change$
    SetNativeTax Z(0)
    AssertEquals "Natives.Tax", "3"
    Natives.Tax := 5
    AssertEquals "Natives.Tax", "5"
    Natives.Tax := 0
    AssertEquals "Natives.Tax", "0"
    Natives.Tax := 50
    AssertEquals "Natives.Tax", "50"
    Assert "Natives.Change$ < h"
    SetNativeTax T
    AssertEquals "Natives.Tax", "T"
  Else
    Assert "IsEmpty(Natives) or IsNum(Natives)"
    Assert "IsEmpty(Natives.Change) or IsString(Natives.Change)"
    Assert "IsEmpty(Natives.Change$) or IsNum(Natives.Change$)"
    Assert "IsEmpty(Natives.Gov) or IsString(Natives.Gov)"
    Assert "IsEmpty(Natives.Gov$) or IsNum(Natives.Gov$)"
    Assert "IsEmpty(Natives.Happy) or IsString(Natives.Happy)"
    Assert "IsEmpty(Natives.Happy$) or IsNum(Natives.Happy$)"
    Assert "IsEmpty(Natives.Race) or IsString(Natives.Race)"
    Assert "IsEmpty(Natives.Race$) or IsNum(Natives.Race$)"
    Assert "IsEmpty(Natives.Tax) or IsNum(Natives.Tax)"
    AssertThrows "SetNativeTax 1"
  EndIf

  % Orbit
  Assert "IsNum(Orbit)"
  Assert "IsNum(Orbit.Enemy)"
  Assert "IsNum(Orbit.Own)"
  AssertEquals "Orbit.Enemy + Orbit.Own", "Orbit"
  AssertEquals "CountShips(Orbit$="&Id&")", "Orbit"
  AssertEquals "CountShips(Loc.X="&Loc.X&" and Loc.Y="&Loc.Y&")", "Orbit"
  AssertEquals "CountShips(Loc.X="&Loc.X&" and Loc.Y="&Loc.Y&" and Owner$=My.Race$)", "Orbit.Own"

  % Owner
  Assert "IsNum(Owner$)"
  If Owner$ Then
    Assert "IsString(Owner)"
    Assert "IsString(Owner.Adj)"
    Assert "Owner$>0 and Owner$<=12"
    AssertEquals "Owner", "Player(Owner$).Race.Short"
    AssertEquals "Owner.Adj", "Player(Owner$).Race.Adj"
  Else
    AssertNull "Owner"
    AssertNull "Owner.Adj"
  EndIf

  % Money, Supplies
  If Owner$=My.Race$ Then
    Assert "IsNum(Money)"
    Assert "IsNum(Supplies)"
    If Supplies > 5 Then
      M := Money
      s := Supplies

      % Sell 3 supplies
      SellSupplies 3
      AssertEquals "Money", "m+3"
      AssertEquals "Supplies", "s-3"
      SellSupplies -3
      AssertEquals "Money", "m"
      AssertEquals "Supplies", "s"

      % Sell too many supplies
      Try SellSupplies (s+1)
      AssertEquals "Money", "m"
      AssertEquals "Supplies", "s"

      % Sell all supplies
      SellSupplies s
      AssertEquals "Money", "m+s"
      AssertEquals "Supplies", "0"
      SellSupplies -s
      AssertEquals "Money", "m"
      AssertEquals "Supplies", "s"

      % Sell too many supplies, with partial sale
      Build.Remainder := 0
      SellSupplies s+1, "n"
      AssertEquals "Money", "m+s"
      AssertEquals "Supplies", "0"
      AssertEquals "Build.Remainder", "1"

      SellSupplies -s
      AssertEquals "Money", "m"
      AssertEquals "Supplies", "s"
    EndIf
  Else
    Assert "IsEmpty(Money) or IsNum(Money)"
    Assert "IsEmpty(Supplies) or IsNum(Supplies)"
    AssertThrows "SellSupplies 1"
  EndIf

  % Temp
  If Owner$=My.Race$ Or Temp
    Assert "IsString(Temp)"
    Assert "IsNum(Temp$)"
    Assert "Temp$>=0 and Temp$<=100"
  Else
    AssertNull "Temp"
    AssertNull "Temp$"
  EndIf

  % Base stuff
  If Owner$=My.Race$ And Base.YesNo Then
    % Damage
    Assert "IsNum(Damage)"
    Assert "Damage >= 0 and Damage <= 100"

    % Mission
    Assert "IsString(Mission)"
    Assert "IsNum(Mission$)"

    M := Mission$
    SetMission 3
    AssertEquals "Mission$", "3"
    SetMission Z(0)
    AssertEquals "Mission$", "3"
    Mission$ := 5
    AssertEquals "Mission$", "5"
    SetMission M
    AssertEquals "Mission$", "M"

    % Fighters
    Assert "IsNum(Fighters)"
    Assert "IsNum(Fighters.Max)"
    % FIXME: this doesn't yet work in PCC2, which cannot yet undo fighter building
    % If Fighters<Fighters.Max And Mined.T>10 And Mined.D>10 And Mined.M>10 And Money>100 Then
    %   Print id
    %   f := Fighters
    %   BuildFighters 1
    %   AssertEquals "Fighters", "f+1"
    %   BuildFighters -1
    %   AssertEquals "Fighters", "f"
    %   Print "."
    % EndIf

    % Build
    Assert "Build.YesNo=0 or Build.YesNo=1"
    Assert "IsNum(Build.Beam$)"
    Assert "IsNum(Build.Beam.Count)"
    Assert "IsNum(Build.Engine$)"
    Assert "IsNum(Build.Torp$)"
    Assert "IsNum(Build.Torp.Count)"
    If Build.YesNo
      Assert "IsString(Build)"
      Assert "IsString(Build.Short)"
      Assert "IsNum(Build.Hull$)"
      AssertEquals "Build", "Hull(Build.Hull$).Name"
      AssertEquals "Build.Short", "Hull(Build.Hull$).Name.Short"
    Else
      AssertNull "Build.Hull$"
      AssertNull "Build"
      AssertNull "Build.Short"
    EndIf
    % FIXME: BuildShip

    % Shipyard
    Assert "IsEmpty(Shipyard) or IsString(Shipyard)"
    Assert "IsEmpty(Shipyard.Action) or IsString(Shipyard.Action)"
    Assert "IsEmpty(Shipyard.Name) or IsString(Shipyard.Name)"
    Assert "IsNum(Shipyard.Id)"
    If IsEmpty(Shipyard) Then
      % We're not fixing any ship, so try whether we can check one.
      % Note that we're checking Shipyard.Name only for v2, because v1
      % attaches the ship Id to the value.
      AssertEquals "Shipyard.Id", "0"
      AssertNull "Shipyard.Action"
      AssertNull "Shipyard.Name"
      If Orbit.Own>0 Then
        % Own ships can be fixed and recycled
        s := FindShip(Ship.Loc.X=Planet.Loc.X And Ship.Loc.Y=Planet.Loc.Y And Owner$=My.Race$)
        Assert "s"

        % Fix, base command
        FixShip s
        AssertEquals "Shipyard.Id", "s"
        AssertEquals "Shipyard.Action", "'Fix'"
        Assert "Not IsEmpty(Shipyard)"
        If IsV2 Then AssertEquals "Shipyard.Name", "Ship(s).Name"

        % Recycle, base command
        RecycleShip s
        AssertEquals "Shipyard.Id", "s"
        AssertEquals "Shipyard.Action", "'Recycle'"
        Assert "Not IsEmpty(Shipyard)"
        If IsV2 Then AssertEquals "Shipyard.Name", "Ship(s).Name"

        % Don't cancel
        FixShip Z(0)
        AssertEquals "Shipyard.Id", "s"
        AssertEquals "Shipyard.Action", "'Recycle'"

        % Cancel using FixShip
        FixShip 0
        AssertEquals "Shipyard.Id", "0"
        AssertNull "Shipyard.Action"
        AssertNull "Shipyard.Name"
        AssertNull "Shipyard"

        % Fix ship command
        With Ship(s) Do FixShip
        AssertEquals "Shipyard.Id", "s"
        AssertEquals "Shipyard.Action", "'Fix'"
        Assert "Not IsEmpty(Shipyard)"
        If IsV2 Then AssertEquals "Shipyard.Name", "Ship(s).Name"

        % Recycle, ship command
        With Ship(s) Do RecycleShip
        AssertEquals "Shipyard.Id", "s"
        AssertEquals "Shipyard.Action", "'Recycle'"
        Assert "Not IsEmpty(Shipyard)"
        If IsV2 Then AssertEquals "Shipyard.Name", "Ship(s).Name"

        % Don't cancel
        RecycleShip Z(0)
        AssertEquals "Shipyard.Id", "s"
        AssertEquals "Shipyard.Action", "'Recycle'"

        % Cancel using RecycleShip
        RecycleShip 0
        AssertEquals "Shipyard.Id", "0"
        AssertNull "Shipyard.Action"
        AssertNull "Shipyard.Name"
        AssertNull "Shipyard"

        Selftest.DidShipyard := BitOr(Selftest.DidShipyard, 1)
      EndIf
      If Orbit.Enemy>0 Then
        % Enemy ships can be fixed
        s := FindShip(Ship.Loc.X=Planet.Loc.X And Ship.Loc.Y=Planet.Loc.Y And Owner$<>My.Race$)
        Assert "s"

        % Fix, base command
        FixShip s
        AssertEquals "Shipyard.Id", "s"
        AssertEquals "Shipyard.Action", "'Fix'"
        Assert "Not IsEmpty(Shipyard)"
        If IsV2 Then AssertEquals "Shipyard.Name", "Ship(s).Name"

        % Recycle, base command
        AssertThrows "RecycleShip s"
        AssertEquals "Shipyard.Id", "s"
        AssertEquals "Shipyard.Action", "'Fix'"
        Assert "Not IsEmpty(Shipyard)"
        If IsV2 Then AssertEquals "Shipyard.Name", "Ship(s).Name"

        % Cancel using FixShip
        FixShip 0
        AssertEquals "Shipyard.Id", "0"
        AssertNull "Shipyard.Action"
        AssertNull "Shipyard.Name"
        AssertNull "Shipyard"

        % Fix ship command
        With Ship(s) Do FixShip
        AssertEquals "Shipyard.Id", "s"
        AssertEquals "Shipyard.Action", "'Fix'"
        Assert "Not IsEmpty(Shipyard)"
        If IsV2 Then AssertEquals "Shipyard.Name", "Ship(s).Name"

        % Recycle, ship command
        With Ship(s) Do AssertThrows "RecycleShip"
        AssertEquals "Shipyard.Id", "s"
        AssertEquals "Shipyard.Action", "'Fix'"
        Assert "Not IsEmpty(Shipyard)"
        If IsV2 Then AssertEquals "Shipyard.Name", "Ship(s).Name"

        % Cancel using RecycleShip
        RecycleShip 0
        AssertEquals "Shipyard.Id", "0"
        AssertNull "Shipyard.Action"
        AssertNull "Shipyard.Name"
        AssertNull "Shipyard"

        Selftest.DidShipyard := BitOr(Selftest.DidShipyard, 2)
      EndIf
    EndIf

    % Storage
    For i:=0 To 11 Do Assert "IsNum(Storage.Ammo(i))"
    For i:=0 To 10 Do Assert "IsNum(Storage.Beams(i))"
    For i:=0 To 9 Do Assert "IsNum(Storage.Engines(i))"
    For i:=0 To Selftest.NumHulls Do Assert "IsNum(Storage.Hulls(i))"
    For i:=0 To 10 Do Assert "IsNum(Storage.Launchers(i))"
    AssertEquals "Storage.Ammo(11)", "Fighters"

    T:=0
    For i:=1 To 11 Do T:=T + Storage.Ammo(i)
    AssertEquals "Storage.Ammo(0)", T

    T:=0
    For i:=1 To 10 Do T:=T + Storage.Beams(i)
    AssertEquals "Storage.Beams(0)", T

    T:=0
    For i:=1 To 9 Do T:=T + Storage.Engines(i)
    AssertEquals "Storage.Engines(0)", T

    T:=0
    For i:=1 To Selftest.NumHulls Do T:=T + Storage.Hulls(i)
    AssertEquals "Storage.Hulls(0)", T

    T:=0
    For i:=1 To 10 Do T:=T + Storage.Launchers(i)
    AssertEquals "Storage.Launchers(0)", T

    AssertNull "Storage.Ammo(-1)"
    AssertNull "Storage.Ammo(12)"
    AssertNull "Storage.Beams(-1)"
    AssertNull "Storage.Beams(11)"
    AssertNull "Storage.Engines(-1)"
    AssertNull "Storage.Engines(10)"
    AssertNull "Storage.Hulls(-1)"
    AssertNull "Storage.Hulls(" & (Selftest.NumHulls+1) & ")"
    AssertNull "Storage.Launchers(-1)"
    AssertNull "Storage.Launchers(11)"
    % FIXME: BuildTorps, BuildBeams, BuildEngines, BuildHulls, BuildLaunchers

    % Tech
    Assert "IsNum(Tech.Beam)"
    Assert "IsNum(Tech.Engine)"
    Assert "IsNum(Tech.Hull)"
    Assert "IsNum(Tech.Torpedo)"
    Assert "Tech.Beam>0 and Tech.Beam<=10"
    Assert "Tech.Engine>0 and Tech.Engine<=10"
    Assert "Tech.Hull>0 and Tech.Hull<=10"
    Assert "Tech.Torpedo>0 and Tech.Torpedo<=10"
    % FIXME: SetTech

    Selftest.DidBase := 1
  Else If IsEmpty(Z(Base.YesNo)) Then
    % Damage
    AssertNull "Damage"

    % Mission
    AssertNull "Mission"
    AssertNull "Mission$"
    AssertThrows "SetMission 1"

    % Fighters
    AssertNull "Fighters"
    AssertNull "Fighters.Max"
    AssertThrows "BuildFighters 1"

    % Build
    AssertNull "Build.YesNo"
    AssertNull "Build.Beam$"
    AssertNull "Build.Beam.Count"
    AssertNull "Build.Engine$"
    AssertNull "Build.Torp$"
    AssertNull "Build.Torp.Count"
    AssertNull "Build"
    AssertNull "Build.Short"
    AssertNull "Build.Hull$"

    % Shipyard
    AssertNull "Shipyard"
    AssertNull "Shipyard.Id"
    AssertNull "Shipyard.Action"
    AssertNull "Shipyard.Name"
    AssertThrows "FixShip 1"
    AssertThrows "RecycleShip 1"

    % Storage
    AssertNull "Storage.Ammo(0)"
    AssertNull "Storage.Beams(0)"
    AssertNull "Storage.Engines(0)"
    AssertNull "Storage.Hulls(0)"
    AssertNull "Storage.Launchers(0)"

    % Tech
    AssertNull "Tech.Beam"
    AssertNull "Tech.Engine"
    AssertNull "Tech.Hull"
    AssertNull "Tech.Torpedo"
    AssertThrows "SetTech 1,5"
  EndIf
Next
ForEach Global.Planet Do Assert "IsNum(Id)"
If Not Selftest.DidOwnPlanet Then Print "Not checking various planet properties, no own planets"
If Not Selftest.DidBuildBase Then Print "Not checking starbase building, no applicable planet"
If Selftest.DidBuildStructures<>15 Then Print "Not checking all structure builds, all planets maxed out (", Selftest.DidBuildStructures, ")"
If Not Selftest.DidBase      Then Print "Not checking base properties, none found"
If Selftest.DidShipyard<>3   Then Print "Not checking all shipyard orders, no applicable ships (", Selftest.DidShipyard, ")"


%
%  Finish
%

Print "Self test finished successfully, ", Selftest.NumTests, " tests."
